unit BOSCompilerU;

interface
uses
  windows,
  sysutils,
  classes,
  contnrs,
  ScannerU,
  PreProcessorU;
  
Procedure CompileBOS( filename : string;
                      EnableBASMExtentions, CompileAndDoNotAssemble, AutoEmitLastRet,ScriptorCompatibility : boolean;
                      const AssemblerExeName, AssemblerParams, AssemblerOutputFile : string );


Type
  ECompileError = class( EScannerError );

  TCaseSensitiveStringList = class( TStringList)
  public
    function Find(const S: string; var Index: Integer): Boolean; override;
    function IndexOf(const S: string): Integer; override;
  end;

  TIdentScope = class;
  TIdent = class;

  TErrorState = (
                 Warning,
                 NormalError,
                 FatalError
                );

  TIdentFlags = (
                 IsPiece,
                 IsLocalVar,
                 IsGlobalVar,
                 IsConst,
                 IsUnitProperty, // not used yet
                 UnknownVarType
                );

  TCodeSnipet = class
  protected
    OldCodeSnipet : TCodeSnipet;
    AsmListing : TStrings;
    fCapturing : Boolean;
    fIdentScope : TIdentScope;
    fControlsAsmListing : Boolean;
    fStartindex : Integer;
    fEndindex : Integer;
    function getitem( Index : Integer ) : string;
    procedure setitem( Index : Integer; const Value : string);
    function getcount : Integer;
  public
    constructor Create( aIdentScope : TIdentScope; aControlsAsmListing : boolean = false );
    destructor Destroy; override;

    procedure Hook( aStartindex : Integer; aEndIndex : Integer = High(integer) );
    procedure UnHook;

    procedure Write;
    procedure WriteTo( aIdentScope : TIdentScope );overload;
    procedure WriteTo( CodeSnipet : TCodeSnipet ); overload;

    property Count : Integer read getcount;
    function Add( const s : string ) : integer;
    procedure Insert( Index : Integer; const s : string );
    procedure Delete( Index : Integer);
    property Items[ Index : Integer ] : string read getitem Write setitem; default;
  end; {TCodeSnipet}

  TIdentScope = class
  protected
    fScopeName : string;
    fParentScope : TIdentScope;
    fIdents : TCaseSensitiveStringList;
    fIdentsOwner : TObjectList;
    fSubScopeCount : integer;
    fLocalVarCount : integer;
    fLabelCount : integer;

    function GetLocalVar( index : Integer ) : TIdent;
    function GetLocalConst( index : Integer ) : TIdent;

  public
    function WriteMsg( const s : string ) : integer; overload;
    Procedure WriteMsg( index : integer; const s : string ) ; overload;

    function WriteStatusMsg( const s : string ) : integer;

    procedure WriteBeginStatusMsg( var index : integer; const s : string );
    function WritePartialStatusMsg( const s : string ) : integer; overload;
    procedure WritePartialStatusMsg( index : integer; const s : string ); overload;
  public
    fAsmListing : TStrings;
    CodeSnipet : TCodeSnipet;

    Inlineable : Boolean;
    IndentLevel : Integer;
    ScopeIndex : integer;
    ExitLabel : string;

    Constructor Create( const aScopeName : string;
                        aParentScope : TIdentScope;
                        aInlineable : Boolean );
    Destructor Destroy; override;

    function IsParent( IdentScope : TIdentScope ) : boolean;

    Function FindIdent( IdentName: string;
                        var IdentNameIsScope : boolean ) : TIdent;
    Function FindSubScope( Const ScopeName : string;
                           var ScopeNameIsIdent : boolean ) : TIdentScope;

    Function NewScope( Const ScopeName : string;
                       Inlineable : boolean ) : TIdentScope;
    Function NewIdent( IdentName : string;
                       aFlags : TIdentFlags ) : TIdent;

    Property ScopeName : string read fScopeName;
    Property ParentScope : TIdentScope read fParentScope;

    property SubScopeCount : Integer read fSubScopeCount;
    property LocalVarCount : Integer read fLocalVarCount;
    property LabelCount : Integer read fLabelCount;

    property LocalVar[ Index : Integer ] : TIdent read GetLocalVar;
    property LocalConst[ Index : Integer ] : TIdent read GetLocalConst;
  end; {TIdentScope}

  TIdent = class
  Protected
    fIdentScope : TIdentScope;
    fName : string;
    fFlags : TIdentFlags;
    fValid : boolean;
    fIndexValue : integer;
    Procedure SetIndexValue( value : integer);
  public
    // the exact meaning of this value changes depending on the ident type
    Property IndexValue : integer read fIndexValue write SetIndexValue;

    Property Valid : boolean read fValid;

    Constructor Create( aIdentScope : TIdentScope;
                        Const aName: string;
                        aFlags : TIdentFlags );


    property IdentScope : TIdentScope read fIdentScope;
    property Name : string read fName;
    property Flags : TIdentFlags read fFlags;
  end; {TIdent}

  TGobalIdentScope = class( TIdentScope )
  public
    GlobalVarCount : integer;
  end; {TGobalIdentScope}

  TBOSCompiler = class
  protected
    Text : TStringList;
    ErrorList : TStringlist;
    Scanner : TScanner;
    PreProcessor : TPreProcessor;
    OutputStream : TStream;
    fAsmListing : TStrings;
    PreProcessorText : TStrings;
  protected
    GlobalIdentScope : TGobalIdentScope;
    CurrentIdentScope : TIdentScope;
    InLineableFunction : TStringList;
    PieceList : TList;

    Function FindIdent( Const IdentName : string;
                        var IdentNameIsScope : boolean ) : TIdent;
    Function NewIdent( Const IdentName : string;
                       Flags : TIdentFlags ) : TIdent;
    Function FindScope( Const ScopeName : string;
                        var ScopeNameIsIdent  : boolean ) : TIdentScope;
    Function ActivateScope( Const ScopeName : string;
                            Inlineable : boolean ) : TIdentScope;
    Procedure DeactivateCurrentScope;

  protected
    function DoConstFactor( Scanner : TScanner ) : Integer;
    function DoConstUnaryFactor( Scanner : TScanner ) : Integer;
    function DoConstMulTerm( Scanner : TScanner ) : Integer;
    function DoConstAddTerm( Scanner : TScanner ) : Integer;
    function DoConstShiftTerm( Scanner : TScanner ) : Integer;
    function DoConstRelationTerm( Scanner : TScanner ) : Integer;
    function DoConstEqualityTerm( Scanner : TScanner ) : Integer;
    function DoConstExpression( Scanner : TScanner ) : Integer;

    function EvaluateConstExpr( expr : string ) : Integer;
  protected
    BASMFilename : string;
    BOSFilename : string;
    fDoOutput : boolean;
    LoopCount : Integer;
    LoopContinueLabel : string;
    UsedBreak : boolean;
    LoopBreakLabel : string;

    LocalVarList : string;

    fEnableBASMExtentions : boolean;
    fAutoEmitLastRet : boolean;
    fScriptorCompatibility : Boolean;

    Procedure ValidatePiece( var aPieceName : string );
    Procedure ValidateAxis( var aAxisName : string );
    Function GetFunc : string;
    Function GetLabel : string;

    Procedure DoPush( value : TIdent );
    Procedure DoPop( value : TIdent );
    Procedure DoPushConst( value : Longint );


    Procedure DoGlobalDeclarations;
    Procedure DoPiece;
    Procedure DoGlobalVar;
    Procedure DoConst;
    Procedure DoFunction;
    Procedure DoInlinableFunction;
    Procedure DoStatementBlock( ForceBlock : boolean = false );
    function DoStatement : boolean;

    Function BiuldParamList( OutputLine : integer ) : integer;

    Procedure DoLocalVar;
    Procedure DoIf;
    Procedure DoWhile;
    Procedure DoDoWhile;
    Procedure DoFor;
    Procedure DoBreak;
    Procedure DoContinue;
    Procedure DoReturn( UseMarker : Boolean );
    Procedure DoBASMInline;

    Procedure AssignmentHelper( OutputLine : integer;
                                var ResultIdent : TIdent); overload;
    Procedure DoAssignment;
    Procedure DoPreIncrement;
    Procedure DoPreDecrement;

    Procedure DoGetUnitProperty( OutputLine : integer );
    Procedure DoSetUnitProperty;
    Procedure DoSleep;
    Procedure DoTurn;
    Procedure DoWaitForTurn;
    Procedure DoMove;
    Procedure DoWaitForMove;
    Procedure DoSpin;
    Procedure DoStopSpin;
    Procedure DoHide;
    Procedure DoShow;
    Procedure DoCache;
    Procedure DoDontCache;
    Procedure DoShade;
    Procedure DoDontShade;
    Procedure DoCallScript( CanReturnValue : boolean{ = false};
                            OutputLine : integer{ = 0};
                            NoCallScriptMarker : boolean{ = False});
    Procedure DoStartScript;
    Procedure DoSignal;
    Procedure DoSetSignalMask;
    Procedure DoExplode;
    Procedure DoEmitSfx;
    Procedure DoRand( OutputLine : integer );
    Procedure DoAttachUnit;
    Procedure DoDropUnit;

    Procedure DoFactor( OutputLine : integer );
    Procedure DoUnaryFactor( OutputLine : integer );
    Procedure DoMulTerm( OutputLine : integer );
    Procedure DoAddTerm( OutputLine : integer );
    Procedure DoShiftTerm( OutputLine : integer );
    Procedure DoRelationTerm( OutputLine : integer );
    Procedure DoEqualityTerm( OutputLine : integer );

    procedure DoBitwiseOr( OutputLine : integer );
    procedure DoBitwiseXor( OutputLine : integer );
    procedure DoBitwiseAnd( OutputLine : integer );
    procedure DoLogicalOr( OutputLine : integer );
    procedure DoLogicalXor( OutputLine : integer );
    procedure DoLogicalAnd( OutputLine : integer );

    Procedure DoExpression( OutputLine : integer );
  protected
    function OnNeedFile( Sender: TPreProcessor;
                         const CallingFilename: string;
                         var FileName, Output: string) : Boolean;
    function OnCommand( Sender: TPreProcessor; const command, param : string) : boolean;
    Procedure OnPositionMapping( Sender : TScanner;
                                 var TokenData : TTokenData );
    function OnLookup(Const Symbol : string; var aToken : TToken) : Boolean;
    Procedure Compile;
  public
    procedure WriteErrorMsg( ErrorState : TErrorState;
                             const Msg : string;
                             RaiseExceptionOnFatalError : boolean = true );
  public
    Constructor Create( aBOSFilename : string;
                        aText : TStringlist;
                        aOutputStream : TStream;
                        aAsmListing : TStrings;
                        aPreProcessorText : TStrings;
                        aEnableBASMExtentions,aAutoEmitLastRet,aScriptorCompatibility : boolean;
                        aBASMFilename : string);
    Destructor destroy; override;

    property ScriptorCompatibility : Boolean read fScriptorCompatibility Write fScriptorCompatibility;
    Property AutoEmitLastRet : boolean read fAutoEmitLastRet write fAutoEmitLastRet;
    Property EnableBASMExtentions : boolean read fEnableBASMExtentions write fEnableBASMExtentions;
    Property DoOutput : boolean read fDoOutput write fDoOutput;
  end; {TBOSCompiler}


Type
  TUnitProperty =record
    Key : Longint;
    Name : string;
    Flags : set of ( SetAble, ArgIsPiece, ArgIsUnitID, IsBool );
    // max of 5
    ArgCount : 0..4;
  end; {TUnitProperty}  
// Indices for set/get value
const
  // set Or get
  Activation         =  1;
  // set Or get
  StandingMoveOrders =  2;
  // set Or get
  StandingFireOrders =  3;
  // get ( 0-100% )
  Health             =  4;
  // set Or get
  InBuildStance      =  5;
  // set Or get ( used by misc. special case missions like transport ships )
  Busy               =  6;
  // get
  Piece_XZ           =  7;
  // get
  Piece_Y            =  8;
  // get
  Unit_XZ            =  9;
  // get
  Unit_Y             = 10;
  // get
  Unit_Height        = 11;
  // get atan of packed x,z coords
  XZ_aTan            = 12;
  // get hypot of packed x,z coords
  XZ_Hypot           = 13;
  // get ordinary two-parameter atan
  aTan               = 14;
  // get ordinary two-parameter hypot
  Hypot              = 15;
  // get
  Ground_Height      = 16;
  // get 0 = unit is built and ready, 1-100 = How much is left to build
  Build_Percent_Left = 17;
  // set Or get ( change which plots we occupy when building opens and closes )
  Yard_Open          = 18;
  // set Or get ( ask other units to clear the area )
  Bugger_Off         = 19;
  // set Or get
  Armored            = 20;

  // various unit properties
  UnitProperties : array [  1..20 ] of TUnitProperty = ( 
                 ( key : Activation;            Name : 'ACTIVATION';
                  Flags : [  SetAble,IsBool ];  ArgCount : 0 ),
                 ( key : StandingMoveOrders;    Name : 'STANDINGMOVEORDERS';
                  Flags : [  SetAble ];         ArgCount : 0 ),
                 ( key : StandingFireOrders;    Name : 'STANDINGFIREORDERS';
                  Flags : [  SetAble ];         ArgCount : 0 ),
                 ( key : Health;                Name : 'HEALTH';
                  Flags : [ ];                  ArgCount : 0 ),
                 ( key : InBuildStance;         Name : 'INBUILDSTANCE';
                  Flags : [  SetAble ];         ArgCount : 0 ),
                 ( key : Busy;                  Name : 'BUSY';
                  Flags : [  SetAble, IsBool ]; ArgCount : 0 ),
                 ( key : Piece_XZ;              Name : 'PIECE_XZ';
                  Flags : [  ArgIsPiece ];      ArgCount : 1 ),
                 ( key : Piece_Y;               Name : 'PIECE_Y';
                  Flags : [  ArgIsPiece ];      ArgCount : 1 ),
                 ( key : Unit_XZ;               Name : 'UNIT_XZ';
                  Flags : [  ArgIsUnitID ];     ArgCount : 1 ),
                 ( key : Unit_Y;                Name : 'UNIT_Y';
                  Flags : [  ArgIsUnitID ];     ArgCount : 1 ),
                 ( key : Unit_Height;           Name : 'UNIT_Height';
                  Flags : [  ArgIsUnitID ];     ArgCount : 1 ),
                 ( key : XZ_aTan;               Name : 'XZ_ATAN';
                  Flags : [ ];                  ArgCount : 1 ),
                 ( key : XZ_Hypot;              Name : 'XZ_HYPOT';
                  Flags : [ ];                  ArgCount : 1 ),
                 ( key : aTan;                  Name : 'ATAN';
                  Flags : [ ];                  ArgCount : 2 ),
                 ( key : Hypot;                 Name : 'HYPOT';
                  Flags : [ ];                  ArgCount : 2 ),
                 ( key : Ground_Height;         Name : 'GROUND_HEIGHT';
                  Flags : [ ];                  ArgCount : 1 ),
                 ( key : Build_Percent_Left;    Name : 'BUILD_PERCENT_LEFT';
                  Flags : [ ];                  ArgCount : 0 ),
                 ( key : Yard_Open;             Name : 'YARD_OPEN';
                  Flags : [  SetAble, IsBool ]; ArgCount : 0 ),
                 ( key : Bugger_Off;            Name : 'BUGGER_OFF';
                  Flags : [  SetAble, IsBool ]; ArgCount : 0 ),
                 ( key : Armored;               Name : 'ARMORED';
                  Flags : [  SetAble, IsBool ]; ArgCount : 0 ) );
                    

implementation
uses
  shellapi,
  FileHandlingU,
  Format_Cob;

// -----------------------------------------------------------------------------
//   TCaseSensitiveStringList
// -----------------------------------------------------------------------------

function TCaseSensitiveStringList.IndexOf(const S: string): Integer;
begin
if not Sorted then
  begin
  for Result := 0 to GetCount - 1 do
    if AnsiCompareStr(Get(Result), S) = 0 then Exit;
  Result := -1;
  end
else if not Find(S, Result) then Result := -1;
end;

function TCaseSensitiveStringList.Find(const S: string; var Index: Integer): Boolean;
var
  L, H, I, C: Integer;
begin
  Result := False;
  L := 0;
  H := Count - 1;
  while L <= H do
  begin
    I := (L + H) shr 1;
    C := AnsiCompareStr(self[I], S);
    if C < 0 then L := I + 1 else
    begin
      H := I - 1;
      if C = 0 then
      begin
        Result := True;
        {if list.Duplicates <> dupAccept then }L := I;
      end;
    end;
  end;
  Index := L;
end; {Find}

// -----------------------------------------------------------------------------

Procedure Exec( const Filename, args : string; const DoRedirect : string );
var
  si : TStartupInfo;
  pi : TProcessInformation;
begin
writeln( 'Attempting to invoke '+Filename);
fillchar(si, sizeof( si ), 0 );
si.cb := sizeof( si );
si.dwFlags := STARTF_USESTDHANDLES;
si.hStdInput := GetStdHandle( STD_INPUT_HANDLE );
si.hStdOutput := GetStdHandle( STD_OUTPUT_HANDLE );
si.hStdError := GetStdHandle( STD_ERROR_HANDLE );
fillchar(pi, sizeof( pi ), 0 );
if CreateProcess( pchar( Filename ), pchar( ' '+args ),
                  nil, nil, true, 0, nil, nil, si, pi) then
  begin
  WaitForSingleObject(pi.hProcess, INFINITE);
  end
else
  begin
  writeln( 'Unable to invoke '+Filename);
  writeln( 'Reason: '+SysErrorMessage( GetLastError ) );
  end;
end; {Exec}

Procedure CompileBOS( filename : string;
                      EnableBASMExtentions, CompileAndDoNotAssemble, AutoEmitLastRet,ScriptorCompatibility : boolean;
                      const AssemblerExeName, AssemblerParams, AssemblerOutputFile : string );
var
  BOSCompiler : TBOSCompiler;
  Stream : TMemoryStream;
  AsmListing, PreProcessorText : TStringlist;
  OutputFileName : string;
  BASMFileName : string;
  PreProcessorFilename : string;
  ErrorFileName : string;
begin
Stream := nil;
AsmListing := nil;
PreProcessorText := nil;
BOSCompiler := nil;
try
  AsmListing := TStringlist.create;
  PreProcessorText := TStringlist.create;
  Stream := TMemoryStream.create;
  filename := FindFile( filename );
  BASMFileName := ChangeFileExt( filename, '.basm');
  PreProcessorFilename := ChangeFileExt( filename, '.prebasm');
  OutputFileName := ChangeFileExt( filename, '.cob');
  ErrorFileName := ChangeFileExt( filename, '_BASM_error.txt');
  writeln( 'Source: '+filename );
  writeln( 'Dest: '+BASMFileName );
  writeln( 'Preprocessor Output: '+PreProcessorFilename );
  writeln( 'BASM error log: '+ErrorFileName);
  try
    BOSCompiler := TBOSCompiler.Create( FileName, Loadfile( filename ),
                                        Stream, AsmListing, PreProcessorText,
                                        EnableBASMExtentions,AutoEmitLastRet,ScriptorCompatibility,BASMFileName );
  finally
    if FileExists( BASMFileName ) then
      DeleteFile( BASMFileName );
    AsmListing.SaveToFile( BASMFileName );
    if FileExists( PreProcessorFilename ) then
      DeleteFile( PreProcessorFilename );
    PreProcessorText.SaveToFile( PreProcessorFilename );
    if BOSCompiler.DoOutput and not CompileAndDoNotAssemble then
      begin
      writeln( 'Assembling BASM to COB' );  
      if FileExists( OutputFileName ) then
        DeleteFile( OutputFileName );
      BASMFileName := AnsiQuotedStr(BASMFileName, '"' );
      OutputFileName := AnsiQuotedStr(OutputFileName, '"' );
      Exec( AssemblerExeName, BASMFileName +' '+ OutputFileName +' ' +
                              AssemblerParams, AssemblerOutputFile);
      If FileExists( 'errors.txt' ) then
        begin
        // and error occured while making the COB, move the 'error.txt' file to
        // the same directory as the rest of the output.
        AsmListing.LoadFromFile( 'errors.txt' );
        DeleteFile( 'errors.txt' );
        if FileExists( ErrorFileName ) then
          DeleteFile( ErrorFileName );
        AsmListing.SaveToFile( ErrorFileName );
        end;
      end
    else if not BOSCompiler.DoOutput then
      begin
      writeln( 'Errors in source' );
      writeln( BOSCompiler.ErrorList.text );
      end;
    writeln( 'Finished Compiling!' );
  end;
finally
  Stream.free;
  AsmListing.free;
  PreProcessorText.free;
  BOSCompiler.free;
end;
end; {CompileBOS}

// -----------------------------------------------------------------------------
//   TCodeSnipet
// -----------------------------------------------------------------------------

constructor TCodeSnipet.Create( aIdentScope : TIdentScope;
                                aControlsAsmListing : Boolean = false );
begin
inherited Create;
fIdentScope := aIdentScope;
Assert( fIdentScope <> nil );
fControlsAsmListing := aControlsAsmListing;
if fControlsAsmListing then
  begin
  fIdentScope.CodeSnipet := Self;
  AsmListing := aIdentScope.fAsmListing;
  end
else
  begin
  AsmListing := TStringList.Create;
  Assert( fIdentScope.CodeSnipet <> nil );
  end;
fEndindex := maxint;
end; {Create}

destructor TCodeSnipet.Destroy;
begin
if not fControlsAsmListing then
  begin
  FreeAndNil( AsmListing );
  if (fIdentScope <> nil) and (fIdentScope.CodeSnipet <> nil) then
    UnHook;
  end
else
  AsmListing := nil;
inherited;
end; {Destroy}

function TCodeSnipet.Getitem( Index : Integer ) : string;
begin
if (Index - fStartindex < 0) or (Index > fEndindex) then
  begin
  if OldCodeSnipet <> nil then
    Result := OldCodeSnipet.GetItem( Index )
  else 
    asm int 3 end;
  end
else
  Result := AsmListing[ Index - fStartindex ]
end; {Getitem}

procedure TCodeSnipet.Setitem( Index : Integer; const Value : string );
begin
if (Index - fStartindex < 0) or (Index > fEndindex) then
  begin
  if OldCodeSnipet <> nil then
    OldCodeSnipet.Setitem( Index, Value )
  else 
    asm int 3 end;
  end
else
  AsmListing[ Index - fStartindex ] := Value;
end; {Setitem}

function TCodeSnipet.GetCount : Integer;
begin
Result := fStartindex+ AsmListing.Count;
end; {GetCount}

function TCodeSnipet.Add( const s : string ) : integer;
begin
Result := fStartindex + AsmListing.Add( s );
end; {Add}

procedure TCodeSnipet.Insert( Index : Integer; const s : string );
begin
if (Index - fStartindex < 0) or (Index > fEndindex) then
  begin
  if OldCodeSnipet <> nil then
    begin
    Inc( fStartindex );
    OldCodeSnipet.Insert( Index, s );
    end
  else
    asm int 3 end;
  end
else
  AsmListing.Insert( Index - fStartindex, s );
end; {Add}

procedure TCodeSnipet.Delete( Index : Integer);
begin
if (Index - fStartindex < 0) or (Index > fEndindex) then
  begin
  if OldCodeSnipet <> nil then
    begin
    Inc( fStartindex );
    OldCodeSnipet.Delete( Index );
    end
  else
    asm int 3 end;
  end
else
  AsmListing.Delete( Index - fStartindex );
end; {Delete}

procedure TCodeSnipet.Hook( aStartindex : Integer; aEndIndex : Integer = High(integer) );
begin
Assert( fStartindex >= 0 );
Assert( fIdentScope <> nil );
Assert( fIdentScope.CodeSnipet <> nil );
if fControlsAsmListing or fCapturing then Exit;
if fIdentScope.CodeSnipet <> self then
  begin
  OldCodeSnipet := fIdentScope.CodeSnipet;
  fIdentScope.CodeSnipet := self;
  fStartindex := aStartindex;
  fEndindex := aEndIndex;
  fCapturing := true;
  end
else
  Assert( False );
end; {Hook}

procedure TCodeSnipet.UnHook;
begin
if fControlsAsmListing or not fCapturing then Exit;
if fIdentScope.CodeSnipet = self then
  begin
  fIdentScope.CodeSnipet := OldCodeSnipet;
  fEndindex := fStartindex + AsmListing.Count-1;
  fCapturing := False;
  end;
end; {UnHook}

procedure TCodeSnipet.Write;
var
  Index : Integer;
begin
if fControlsAsmListing or (AsmListing.Count = 0) then Exit;
for Index := 0 to AsmListing.Count-1 do
  fIdentScope.CodeSnipet.Add( AsmListing[ Index ] );
end; {Write}

procedure TCodeSnipet.WriteTo( aIdentScope : TIdentScope );
var
  Index : Integer;
begin
if (AsmListing.Count = 0) or (aIdentScope = fIdentScope) then Exit;
for Index := 0 to AsmListing.Count-1 do
  aIdentScope.CodeSnipet.add( AsmListing[ Index ] );
end; {WriteTo}

procedure TCodeSnipet.WriteTo( CodeSnipet : TCodeSnipet );
var
  Index : Integer;
begin
if (AsmListing.Count = 0) or (CodeSnipet = self) then Exit;
for Index := 0 to AsmListing.Count-1 do
  CodeSnipet.add( AsmListing[ Index ] );
end; {WriteTo}

// -----------------------------------------------------------------------------
//   TIdentScope
// -----------------------------------------------------------------------------

Constructor TIdentScope.Create( const aScopeName : string;
                                aParentScope : TIdentScope;
                                aInlineable : Boolean );
begin
inherited create;
assert( aScopeName <> '' );
fScopeName := aScopeName;
fParentScope := aParentScope;
fIdents := TCaseSensitiveStringList.create;
fIdents.Sorted := true;
fIdents.Duplicates := dupError;
fIdentsOwner := TObjectList.create( true );
Inlineable := aInlineable;
fAsmListing := TStringList.Create;
CodeSnipet := TCodeSnipet.Create( Self, true );
end; {Create}

Destructor TIdentScope.Destroy;
begin
FreeAndNil( fAsmListing );
FreeAndNil( CodeSnipet );
FreeAndNil( fIdents );
FreeAndNil( fIdentsOwner );
inherited;
end; {Destroy}

Function TIdentScope.FindIdent( IdentName: string;
                                var IdentNameIsScope : boolean ) : TIdent;
var
  obj : TObject;
  index : integer;
begin
assert( self <> nil );
assert( fIdents <> nil );
result := nil;
IdentNameIsScope := false;
if IdentName = '' then exit;
if fIdents.Find( IdentName, index ) or
   ( Inlineable and
     fIdents.Find( ScopeName+ '_'+IdentName, index ) ) then
  begin
  obj := fIdents.Objects[ index];
  if obj is TIdent then
    result := TIdent( obj )
  else if obj is TIdentScope then
    IdentNameIsScope := true;
  end
else if ParentScope <> nil then
  result := ParentScope.FindIdent( IdentName, IdentNameIsScope );
end; {FindIdent}

function TIdentScope.IsParent( IdentScope : TIdentScope ) : Boolean;
var
  walker : TIdentScope;
begin
walker := IdentScope;
while walker <> nil do
  begin
  if walker = Self then
    begin
    Result := True;
    Exit;
    end;
  walker := walker.ParentScope;
  end;
Result := false;
end; {IsParent}

Function TIdentScope.FindSubScope( Const ScopeName: string;
                                   var ScopeNameIsIdent : boolean ) : TIdentScope;
var
  obj : TObject;
  index : integer;
begin
assert( self <> nil );
assert( fIdents <> nil );
result := nil;
ScopeNameIsIdent := false;
if ScopeName = '' then exit;
if fIdents.Find( ScopeName, index ) then
  begin
  obj := fIdents.Objects[ index];
  if obj is TIdentScope then
    result := TIdentScope( obj )
  else if obj is TIdent then
    ScopeNameIsIdent := true;
  end;
if ( result = nil ) and ( ParentScope <> nil ) then
  Result := ParentScope.FindSubScope( ScopeName, ScopeNameIsIdent );
end; {FindSubScope}

Function TIdentScope.NewScope( const ScopeName : string;
                               Inlineable : boolean ) : TIdentScope;
begin
assert( self <> nil );
assert( fIdents <> nil );
assert( fIdentsOwner <> nil );
if fIdents.IndexOf( ScopeName ) = -1 then
  begin
  result := TIdentScope.create( ScopeName, Self, Inlineable );
  fIdentsOwner.add( result );
  fIdents.AddObject( ScopeName ,result );
  result.ScopeIndex := SubScopeCount;
  inc( fSubScopeCount );
  end
else raise ECompileError.create( 'Redeclared Identifier: "' + ScopeName + '"' );
end; {NewScope}

Function TIdentScope.NewIdent( IdentName : string;
                               aFlags : TIdentFlags ) : TIdent;
begin
assert( self <> nil );
assert( fIdents <> nil );
assert( fIdentsOwner <> nil );
// name mangle
if Inlineable then
  IdentName := ScopeName+ '_'+IdentName;
if fIdents.IndexOf( IdentName ) = -1 then
  begin
  result := TIdent.create( self, IdentName, aFlags );
  fIdentsOwner.add( result );
  fIdents.AddObject( IdentName ,result );
  end
else raise ECompileError.create( 'Redeclared Identifier: "' + IdentName + '"' );
end; {NewIdent}

function TIdentScope.GetLocalConst( index : Integer ) : TIdent;
var
  i : Integer;
  tmp : TObject;
begin
for i := 0 to fIdents.Count -1 do
  begin
  tmp := TObject(fIdents.objects[i]);
  if (tmp is TIdent) and
     (TIdent(tmp).Flags = IsConst) and
     (TIdent(tmp).IndexValue = Index ) then
    begin
    Result := TIdent(tmp);
    Exit;
    end;
  end;
Result := nil;
end; {GetLocalConst}

function TIdentScope.GetLocalVar( index : Integer ) : TIdent;
var
  i : Integer;
  tmp : TObject;
begin
for i := 0 to fIdents.Count -1 do
  begin
  tmp := TObject(fIdents.objects[i]);
  if (tmp is TIdent) and
     (TIdent(tmp).Flags = IsLocalVar) and
     (TIdent(tmp).IndexValue = Index ) then
    begin
    Result := TIdent(tmp);
    Exit;
    end;
  end;
Result := nil;
end; {GetLocalVar}

Procedure TIdentScope.WriteMsg( index : integer; const s : string ) ;
begin
assert( self <> nil );
//writeln( s );
CodeSnipet.Insert( index, s );
end; {WriteMsg}

function TIdentScope.WriteMsg( const s : string ) : integer ;
begin
assert( self <> nil );
//writeln( s );
Result := CodeSnipet.Add( StringOfChar(' ',IndentLevel) + s );
end; {WriteMsg}

function TIdentScope.WriteStatusMsg( const s : string ) : integer;
begin
assert( self <> nil );
result := WriteMsg( '//'+s );
end; {WriteStatusMsg}

Procedure TIdentScope.WriteBeginStatusMsg(  var index : integer; const s : string );
begin
assert( self <> nil );
index := CodeSnipet.add( StringOfChar(' ',IndentLevel) + '// ' + s );
end; {WriteStatusMsg}

Function TIdentScope.WritePartialStatusMsg( const s : string ) : integer;
begin
assert( self <> nil );
result := CodeSnipet.count-1;
CodeSnipet[ result ] := CodeSnipet[ result ] + s;
end; {WriteStatusMsg}

Procedure TIdentScope.WritePartialStatusMsg( index : integer; const s : string );
begin
assert( self <> nil );
Assert( CodeSnipet <> nil );
CodeSnipet[ index ] := CodeSnipet[ index ] + s;
end; {WriteStatusMsg}

// -----------------------------------------------------------------------------
//   TIdent
// -----------------------------------------------------------------------------

Constructor TIdent.Create( aIdentScope : TIdentScope;
                           Const aName: string;
                           aFlags : TIdentFlags );
begin
assert( aIdentScope <> nil );
assert( aName <> '' );
fIdentScope := aIdentScope;
fName := aName;
fFlags := aFlags;
IndexValue := -1;
end; {Create}

Procedure TIdent.SetIndexValue( value : integer);
begin
fIndexValue := value;
fValid := true;
end; {SetIndexValue}

// -----------------------------------------------------------------------------
//   TBOSCompiler
// -----------------------------------------------------------------------------

Constructor TBOSCompiler.Create(  aBOSFilename : string;
                                  aText : TStringlist;
                                  aOutputStream : TStream;
                                  aAsmListing,aPreProcessorText : TStrings;
                                  aEnableBASMExtentions, aAutoEmitLastRet,
                                  aScriptorCompatibility : boolean;
                                  aBASMFilename : string);
begin
Inherited Create;
Text := aText;
OutputStream := aOutputStream;
fAsmListing := aAsmListing;
PreProcessorText := aPreProcessorText;
BASMFilename := aBASMFilename;
BOSFilename := aBOSFilename;
assert( Text <> nil );
assert( OutputStream <> nil );
assert( fAsmListing <> nil );
assert( PreProcessorText <> nil );
EnableBASMExtentions := aEnableBASMExtentions;
AutoEmitLastRet := aAutoEmitLastRet;
ScriptorCompatibility := aScriptorCompatibility;
PieceList := TList.create;
GlobalIdentScope := TGobalIdentScope.create( 'global scope', nil, false );

GlobalIdentScope.NewIdent( 'TRUE', IsConst ).IndexValue := 1;
GlobalIdentScope.NewIdent( 'FALSE', IsConst ).IndexValue := 0;
GlobalIdentScope.NewIdent( 'true', IsConst ).IndexValue := 1;
GlobalIdentScope.NewIdent( 'false', IsConst ).IndexValue := 0;
GlobalIdentScope.NewIdent( 'x-axis', IsConst ).IndexValue := 0;
GlobalIdentScope.NewIdent( 'y-axis', IsConst ).IndexValue := 1;
GlobalIdentScope.NewIdent( 'z-axis', IsConst ).IndexValue := 2;

CurrentIdentScope := GlobalIdentScope;
ErrorList := TStringlist.Create;
Preprocessor := TPreProcessor.create;
Preprocessor.OnNeedFile := OnNeedFile;
Preprocessor.OnCommand := OnCommand;

try
  Compile;
finally
  fAsmListing.Assign( GlobalIdentScope.fAsmListing );
end;
//aText.Text := Scanner.Text;
//aText.SaveToFile('test2.txt');
end; {Create}

Destructor TBOSCompiler.Destroy;
begin
Scanner.free;
GlobalIdentScope.free;
PieceList.free;
ErrorList.free;
Text.free;
Inherited Destroy;
end; {Destroy}

// -----------------------------------------------------------------------------

Procedure TBOSCompiler.WriteErrorMsg( ErrorState : TErrorState;
                                      const Msg : string;
                                      RaiseExceptionOnFatalError : boolean = true ) ;
var
  ErrorType, s : string;
begin
DoOutput := false;
ErrorType := '';
case ErrorState of
  Warning     : ErrorType := 'Warning';
  NormalError : ErrorType := 'Error';
  FatalError  : ErrorType := 'Fatal Error';
end;
if (scanner = nil) or (Preprocessor = nil) then
  s := format( '[%s] : %s', [ ErrorType, Msg] )
else
  s := Format( '[%s] "%s"(%d:%d) : %s', [ ErrorType, scanner.Filename, scanner.Line, scanner.Column, Msg ] );
ErrorList.add( s );
if (ErrorState = FatalError) and RaiseExceptionOnFatalError then
  raise ECompileError.create( s );
end; {WriteErrorMsg}

// -----------------------------------------------------------------------------

// Parse and Translate a Math Factor
function TBOSCompiler.DoConstFactor( Scanner : TScanner ) : integer;
var
  Ident : TIdent;
  IntValue : Int64;
  IdentNameIsScope : boolean;
begin
case Scanner.Token of
  CSTI_OpenRound :
    begin
    Scanner.Match( CSTI_OpenRound );
    Result := DoConstExpression( Scanner );
    Scanner.Match( CSTI_CloseRound );
    end;
  CSTI_Identifier :
    begin
    Ident := FindIdent( Scanner.Value, IdentNameIsScope );
    if Ident = nil then
       begin  // referencing a non-existant variable
       WriteErrorMsg( NormalError, 'Undeclaired identifier, ' + Scanner.Value );
       Ident := NewIdent( Scanner.Value, UnknownVarType );
       end;
    if Ident.Flags in [IsLocalVar, IsGlobalVar ] then
      WriteErrorMsg( FatalError, 'Constant expression expected, but found a varaible '+Ident.Name );
    if not Ident.Valid then
      WriteErrorMsg( FatalError, 'Undeclaired identifier, ' + Ident.Name );
    result := Ident.IndexValue;
    Scanner.GetToken;
    end;
  CSTI_Integer, CSTI_HexInt :
    begin
    IntValue := StrToInt64( Scanner.Value );
    if IntValue > High(Longword) then
      IntValue := IntValue mod High(Longword);
    result := IntValue;
    Scanner.GetToken;
    end;
  CSTI_Real :
    begin
    WriteErrorMsg( NormalError, 'Floatpoint values are not permitted in arbitrary locations ' + Scanner.Value );
    Scanner.GetToken;
    result := 0;
    end
  else Raise ECompileError.create( 'Unknown token: ' + TokenName( Scanner.Token ) );
end;    
end; {DoConstFactor}

function TBOSCompiler.DoConstUnaryFactor( Scanner : TScanner ) : integer;
begin
case Scanner.Token of
  CSTI_BitwiseNot :
    begin
    Scanner.GetToken;
    result := not DoConstFactor( Scanner );
    end;
  CSTI_LogicalNot :
    begin
    Scanner.GetToken;
    result := Integer( not Boolean( DoConstFactor( Scanner ) ) );
    end;
  CSTI_Minus :
    begin
    Scanner.GetToken;
    result := - DoConstFactor( Scanner );
    end;
else result := DoConstFactor( Scanner );
end;
end; {DoConstUnaryFactor}

function TBOSCompiler.DoConstMulTerm( Scanner : TScanner ) : integer;
begin
result := DoConstUnaryFactor( Scanner );
while true do
case Scanner.Token of
  CSTI_Multiply :
    begin
    Scanner.Match( CSTI_Multiply );
    result := result * DoConstUnaryFactor( Scanner );
    end;
  CSTI_Divide :
    begin
    Scanner.Match( CSTI_Divide );
    result := result div DoConstUnaryFactor( Scanner );
    end;
  else break;
end;
end; {DoMulTerm}

function TBOSCompiler.DoConstAddTerm( Scanner : TScanner ) : integer;
begin
result := DoConstMulTerm( Scanner );
while true do
case Scanner.Token of
  CSTI_Plus :
    begin
    Scanner.Match( CSTI_Plus );
    result := result + DoConstMulTerm( Scanner );
    end;
  CSTI_Minus :
    begin
    Scanner.Match( CSTI_Minus );
    result := result - DoConstMulTerm( Scanner );
    end;
  else break;
end;
end; {DoConstAddTerm}

function sar(a : integer; b : byte) : integer;
asm
  mov cl, b;
  sar a, cl;
end;

function TBOSCompiler.DoConstShiftTerm( Scanner : TScanner ) : integer;
begin
Result := DoConstAddTerm( Scanner );
if EnableBASMExtentions then
while true do
case Scanner.Token of
  CSTI_ShiftLeft :
    begin
    Scanner.Match( CSTI_ShiftLeft );
    Result := Result shl DoConstAddTerm( Scanner );
    end;
  CSTI_ShiftRight :
    begin
    Scanner.Match( CSTI_ShiftRight );
    Result := sar( Result, DoConstAddTerm( Scanner ) );
    end;
  else break;
end;
end; {DoConstShiftTerm}

function TBOSCompiler.DoConstRelationTerm( Scanner : TScanner ) : integer;
begin
result := DoConstShiftTerm( Scanner );
case Scanner.Token of
  CSTI_Greater :
    begin
    Scanner.Match( CSTI_Greater );
    result := Integer( result > DoConstShiftTerm( Scanner ) );
    end;
  CSTI_GreaterEqual :
    begin
    Scanner.Match( CSTI_GreaterEqual );
    result := Integer( result >= DoConstShiftTerm( Scanner ) );
    end;
  CSTI_Less :
    begin
    Scanner.Match( CSTI_Less );
    result := Integer( result < DoConstShiftTerm( Scanner ) );
    end;
  CSTI_LessEqual :
    begin
    Scanner.Match( CSTI_LessEqual );
    result := Integer( result <= DoConstShiftTerm( Scanner ) );
    end;
end;
end; {DoConstRelationTerm}

function TBOSCompiler.DoConstEqualityTerm( Scanner : TScanner ) : integer;
begin
result := DoConstRelationTerm( Scanner );
case Scanner.Token of
  CSTI_Equal :
    begin
    Scanner.Match( CSTI_Equal );
    result := Integer( result = DoConstRelationTerm( Scanner ) );
    end;
  CSTI_NotEqual :
    begin
    Scanner.Match( CSTI_Equal );
    result := Integer( result <> DoConstRelationTerm( Scanner ) );
    end;
end;
end; {DoConstEqualityTerm}

function TBOSCompiler.DoConstExpression( Scanner : TScanner ) : integer;
begin
result := DoConstEqualityTerm( Scanner );
while true do
  begin
  if Scanner.token = CSTI_EOF then
    break;
  if Scanner.IfMatch( CSTI_LogicalOr ) then
    result := Integer( boolean( result ) or boolean( DoConstEqualityTerm( Scanner ) ) )
  else if Scanner.IfMatch( CSTI_LogicalXor ) then
    result := Integer( boolean( result ) xor boolean( DoConstEqualityTerm( Scanner  ) ) )
  else if Scanner.IfMatch( CSTI_LogicalAnd ) then
    result := Integer( boolean( result ) and boolean( DoConstEqualityTerm( Scanner  ) ) )
  else if Scanner.IfMatch( CSTI_BitwiseOr ) then
    result := result or DoConstEqualityTerm( Scanner  )
  else if Scanner.IfMatch( CSTI_BitwiseXor ) then
    result := result xor DoConstEqualityTerm( Scanner  )
  else if Scanner.IfMatch( CSTI_BitwiseAnd ) then
    result := result and DoConstEqualityTerm( Scanner  )
  else break;
  end; 
end; {DoConstExpression}

function TBOSCompiler.EvaluateConstExpr( expr : string ) : Integer;
var
  aScanner : TScanner;
  Preprocessor : TPreProcessor;
begin
Preprocessor := TPreProcessor.create;
try
  Preprocessor.OnCommand := OnCommand;
//  Preprocessor.OnNeedFile := OnNeedFile;
  Preprocessor.MainFileData := expr;
  Preprocessor.PreProcess( '', expr );
finally
  Preprocessor.free;
end;
aScanner := TScanner.create( nil, OnLookup, expr );
try
  Result := DoConstExpression( aScanner );
finally
  aScanner.free;
end;
end; {EvaluateConstExpr}

// -----------------------------------------------------------------------------

Function TBOSCompiler.FindIdent( Const IdentName : string;
                                 var IdentNameIsScope : boolean ) : TIdent;
begin
assert( self <> nil );
assert( CurrentIdentScope <> nil );
result := CurrentIdentScope.FindIdent( IdentName, IdentNameIsScope )
end; {FindIdent}

Function TBOSCompiler.NewIdent( Const IdentName : string;
                                Flags : TIdentFlags ) : TIdent;
begin
assert( self <> nil );
assert( CurrentIdentScope <> nil );
result := CurrentIdentScope.NewIdent( IdentName, Flags );
if Flags = IsPiece then
  result.IndexValue := PieceList.add( result )
else if Flags = IsLocalVar then
  begin
  result.IndexValue := CurrentIdentScope.LocalVarCount;
  inc( CurrentIdentScope.fLocalVarCount );
  LocalVarList := LocalVarList + IdentName + ', ';
  end
else if Flags = IsGlobalVar then
  begin
  result.IndexValue := GlobalIdentScope.GlobalVarCount;
  inc( GlobalIdentScope.GlobalVarCount );
  end
end; {NewIdent}

Function TBOSCompiler.FindScope( Const ScopeName : string;
                                 var ScopeNameIsIdent  : boolean ) : TIdentScope;
begin
assert( self <> nil );
assert( CurrentIdentScope <> nil );
Result := CurrentIdentScope.FindSubScope( ScopeName, ScopeNameIsIdent );
end; {FindScope}

Function TBOSCompiler.ActivateScope( Const ScopeName : string;
                                     Inlineable : boolean ) : TIdentScope;
var
  ScopeNameIsIdent : boolean;
begin
assert( self <> nil );
Result := FindScope( ScopeName, ScopeNameIsIdent );
if Result = nil then
  begin
  if ScopeNameIsIdent then
    Raise ECompileError.create( 'Function name "'+ScopeName+'" is used by a variable ' );
  result := CurrentIdentScope.NewScope( ScopeName, Inlineable );
  if Result = nil then
    Raise ECompileError.create( 'Error: Unable to activate scope:' + ScopeName );
  end;
if (result <> nil) then
  CurrentIdentScope := result;
end; {ActivateScope}

Procedure TBOSCompiler.DeactivateCurrentScope;
begin
assert( self <> nil );
assert( CurrentIdentScope <> nil );
if CurrentIdentScope.ParentScope <> nil then
  CurrentIdentScope := CurrentIdentScope.ParentScope
else
  CurrentIdentScope := GlobalIdentScope;
end; {DeactivateCurrentScope}

// -----------------------------------------------------------------------------

function TBOSCompiler.OnNeedFile( Sender: TPreProcessor;
                                  const CallingFilename: string;
                                  var FileName, Output: string) : Boolean;
var
  includefile : TStringlist;
  fullname : string;
begin
fullname := FindFile( FileName );
if fullname <> '' then
  begin
  includefile := Loadfile( fullname );
  try
    Output := includefile.text;
  finally
    includefile.free;
  end;
  result := true;
  end
else
  result := false;
end; {OnNeedFile}

function TBOSCompiler.OnCommand( Sender: TPreProcessor;
                                 const command, param : string) : boolean;
begin
if Command = 'if' then
  begin
  result := true;
  Sender.DefineState.Add.DoWrite := EvaluateConstExpr( param ) <> 0;
  end
else
  result := false;  
end; {OnIf}

Procedure TBOSCompiler.OnPositionMapping( Sender : TScanner;
                                          var TokenData : TTokenData );
var
  Res : TLineInfoResults;
begin
if Preprocessor.CurrentLineInfo.GetLineInfo( TokenData.Position-1, Res ) then
  begin
  TokenData.Filename := Res.Name;
  TokenData.Line := Res.Row;
  TokenData.Column := Res.Column;
  TokenData.Position := Res.Position;
  end;
end; {OnPositionMapping}

function TBOSCompiler.OnLookup(Const Symbol : string; var aToken : TToken) : Boolean;
var a : boolean;
begin
if (CurrentIdentScope <> nil) and
   (CurrentIdentScope.FindIdent( Symbol, a ) <> nil) then
  begin
  Result := True;
  aToken := CSTI_Identifier;
  end
else
  Result := False;
end; {OnLookup}

Procedure TBOSCompiler.Compile;
var
  StartTime,EndTime,freq : Int64;
  CompileTime : Extended;
  index : integer;
  data : string;
begin
assert( self <> nil );
ErrorList.clear;
QueryPerformanceFrequency( freq );
QueryPerformanceCounter( StartTime );
try try
  // do pre-processing

  Preprocessor.Clear;
  Preprocessor.MainFileName := BOSFilename;
  Preprocessor.MainFileData := Text.Text;
  try
    Preprocessor.PreProcess( BOSFilename, data );
  finally
    PreProcessorText.Text := data;
  end;

  // construct the Lexicon scanner
  if Scanner <> nil then
    FreeAndNil( Scanner );
  Scanner := TScanner.create( OnPositionMapping , OnLookup, data );
  // do the parsing
  LoopCount := 0;
  LoopContinueLabel := '';
  LoopBreakLabel := '';
  UsedBreak := false;
  DoOutput := true;
  // parse the unit script
  DoGlobalDeclarations;
  // create the script output
except
  On E : Exception do
    WriteErrorMsg( FatalError, E.message, false  );
end;
finally
  QueryPerformanceCounter( EndTime );
  CompileTime := ( EndTime - StartTime ) / freq;
  CurrenTIdentScope.WriteStatusMsg( 'Compile time  : ' + FloatToStrF( CompileTime,ffExponent,2,0 ) + ' seconds.' );
  DoOutput := ErrorList.count = 0;
  if not DoOutput then
    begin
    WriteErrorMsg( FatalError, 'Could not compile dependant file.', false );
    CurrenTIdentScope.WriteStatusMsg( 'Errors in source code' );
    for index := 0 to ErrorList.count-1 do
      CurrentIdentScope.WriteStatusMsg( ErrorList[index] );
    end;
end;
end; {Compile}

Procedure TBOSCompiler.ValidatePiece( var aPieceName : string );
var
  ident : TIdent;
  IdentNameIsScope : boolean;
begin
aPieceName := Scanner.Value;
if Scanner.Token in [ CSTI_Integer,CSTI_HexInt] then
  begin
  end
else if Scanner.Token = CSTI_Identifier then
  begin
  ident := FindIdent( Scanner.Value, IdentNameIsScope );
  if Ident = nil then
     begin
     WriteErrorMsg( NormalError, 'Undeclaired identifier, ' + Scanner.Value );
     ident := NewIdent( Scanner.Value, UnknownVarType );
     end;
  assert( Ident <> nil );
  if (ident.Flags <> IsPiece) then
    WriteErrorMsg( NormalError, 'Expecting piece name, but found a unknown variable type' )
  end
else
  WriteErrorMsg( NormalError, 'Expecting piece name or value, but found a ' + TokenName( Scanner.Token ) );
Scanner.GetToken;
end; {ValidatePiece}

Procedure TBOSCompiler.ValidateAxis( var aAxisName : string );
var
  Ident : TIdent;
  IdentNameIsCope : Boolean;
begin
aAxisName := Scanner.Value;
if Scanner.Token in [ CSTI_Integer,CSTI_HexInt] then
  begin
  end
else if Scanner.Token = CSTI_Identifier then
  begin
  Ident := CurrentIdentScope.FindIdent( aAxisName, IdentNameIsCope );
  if not IdentNameIsCope and
    (Ident <> nil) and Ident.Valid then
    begin
//    Result := Ident.IndexValue
    end
  else
    WriteErrorMsg( NormalError, 'Expecting axis name or value, but found ' + aAxisName )
  end
else
  WriteErrorMsg( NormalError, 'Expecting axis name or value, but found a ' + TokenName( Scanner.Token ) );
Scanner.GetToken;
end; {ValidateAxis}

Function TBOSCompiler.GetFunc : string;
var
  Scope : TIdentScope;
  ScopeNameIsIdent : boolean;
begin
if Scanner.Token = CSTI_Identifier then
  begin
  Scope := CurrentIdentScope.FindSubScope( Scanner.Value, ScopeNameIsIdent );
  if (Scope = nil) then
    begin
    if ScopeNameIsIdent then
      WriteErrorMsg( FatalError, 'Function name "'+Scanner.Value+'" is used by a variable ' )
    else
      WriteErrorMsg( FatalError, 'Undeclaired function name, ' + Scanner.Value );
    end;  
  result := Scope.ScopeName;
  Scanner.GetToken;
  end
else
  begin
  result := Scanner.Value;
  WriteErrorMsg( NormalError, 'Expecting function name, but found a ' + TokenName( Scanner.Token ) );
  end;
end; {GetFunc}

Function TBOSCompiler.GetLabel : string;
begin
if CurrentIdentScope <> GlobalIdentScope then
  begin
  result := CurrentIdentScope.ScopeName+'_'+IntToStr( CurrentIdentScope.LabelCount );
  inc( CurrentIdentScope.fLabelCount );
  end
else
  begin
  result := '';
  WriteErrorMsg( NormalError, 'Labels are not allowed in the global scope' );
  end;
end; {GetLabel}

Procedure TBOSCompiler.DoPush( value : TIdent );
begin
assert( value <> nil );
case value.Flags of
  IsLocalVar :
    begin
    CurrentIdentScope.WriteMsg( 'pushl ' + value.Name + ';');
    end;
  IsGlobalVar :
    begin
    CurrentIdentScope.WriteMsg( 'pushs ' + value.Name + ';' );
    end;
  IsPiece :
    begin
    CurrentIdentScope.WriteMsg( 'pushc ' + IntTostr( value.IndexValue ) + '; // '+ value.Name );
    end; // todo
  IsConst :
    begin
    CurrentIdentScope.WriteMsg( 'pushc ' + IntTostr( value.IndexValue ) + '; // '+ value.Name );
    end; // todo
  else
    begin
    WriteErrorMsg( NormalError, 'Unknown result type, '+value.Name );
    CurrentIdentScope.WriteMsg( 'push ' + value.Name + ';' );
    end;
end;
end; {DoPush}

Procedure TBOSCompiler.DoPop( value : TIdent );
begin
assert( value <> nil );
case value.Flags of
  IsLocalVar :
    begin
    CurrentIdentScope.WriteMsg( 'popl ' + value.Name + ';');
    end;
  IsGlobalVar :
    begin
    CurrentIdentScope.WriteMsg( 'pops ' + value.Name + ';' );
    end;
  IsPiece,
  IsConst :
    begin
    WriteErrorMsg( NormalError, 'Can not assign to a read-only identifier, ' + value.Name );
    CurrentIdentScope.WriteMsg( '// pops ' + value.Name + ';' );
    end;
  else
    begin
    WriteErrorMsg( NormalError, 'Unknown result type, or result is readonly' );
    CurrentIdentScope.WriteMsg( 'pop ' + value.Name + ';' );
    end;
end;
end; {DoPop}

Procedure TBOSCompiler.DoPushConst( value : Longint );
begin
CurrentIdentScope.WriteMsg( 'pushc ' + IntToStr( value ) + ';' );
end; {DoPushConst}

Procedure TBOSCompiler.DoGlobalDeclarations;
begin
assert( self <> nil );
while true do
  case Scanner.Token of
    CSTII_piece : DoPiece;
    CSTII_GlobalVar : DoGlobalVar;
    CSTI_Identifier : DoFunction;
    CSTI_Inlinable : DoInlinableFunction;
    CSTII_const : DoConst;
  else break;
  end;
end; {DoGlobalDeclarations}

Procedure TBOSCompiler.DoConst;
var
  NamedConstant : TIdent;
  ConstList : string;
begin
Scanner.Match( CSTII_const );
ConstList := 'const ';
try
  repeat
    if Scanner.Token <> CSTI_Identifier then
      Scanner.Match( CSTI_Identifier );
    NamedConstant := NewIdent( Scanner.Value, IsConst );
    assert( NamedConstant <> nil );
    ConstList := ConstList + Scanner.Value + ' ';
    Scanner.GetToken;
    Scanner.Match( CSTI_Assignment );
    ConstList := ConstList + '= ';
    NamedConstant.IndexValue := DoConstExpression( Scanner );
    ConstList := ConstList + IntToStr( NamedConstant.IndexValue) + ', ';
  until not Scanner.IfMatch( CSTI_Comma );
  setlength( ConstList, length( ConstList )-2 );
  Scanner.Match( CSTI_SemiColon );
  ConstList := ConstList +';';  
finally
  CurrentIdentScope.WriteStatusMsg( ConstList );
end;
end; {DoConst}

Procedure TBOSCompiler.DoPiece;
var
  PieceList : string;
begin
Scanner.Match( CSTII_piece );
PieceList := 'piece ';
try
  repeat
    if Scanner.Token <> CSTI_Identifier then
      Scanner.Match( CSTI_Identifier );
    NewIdent( Scanner.Value, IsPiece );
    PieceList := PieceList + Scanner.Value + ', ';
    Scanner.GetToken;
  until not Scanner.IfMatch( CSTI_Comma );
  setlength( PieceList, length( PieceList )-2 );
  Scanner.Match( CSTI_SemiColon );
  PieceList := PieceList + ';';
finally
  CurrentIdentScope.WriteMsg( PieceList );
end;
end; {DoPiece}

Procedure TBOSCompiler.DoGlobalVar;
var
  VarList : string;
begin
Scanner.Match( CSTII_GlobalVar );
VarList := 'static-var ';
try
  repeat
    if Scanner.Token <> CSTI_Identifier then
      Scanner.Match( CSTI_Identifier );
    NewIdent( Scanner.Value,IsGlobalVar );
    VarList := VarList + Scanner.Value + ', ';
    Scanner.GetToken;
  until not Scanner.IfMatch( CSTI_Comma );
  setlength( VarList, length( VarList )-2 );  
  Scanner.Match( CSTI_SemiColon );
  VarList := VarList + ';';
finally
  CurrentIdentScope.WriteMsg( VarList );
end;
end; {DoGlobalVar}

Procedure TBOSCompiler.DoInlinableFunction;
var
  FunctionName : string;
  ReturnsData : Boolean;
  outputline : integer;
  StartScope : TIdentScope;
begin
Scanner.Match( CSTI_Inlinable );
StartScope := CurrentIdentScope;
StartScope.WriteBeginStatusMsg( outputline, 'inline ' );
ReturnsData := Scanner.IfMatch( CSTI_Returns );
if ReturnsData then
  StartScope.WritePartialStatusMsg( outputline, 'returns ');

FunctionName := Scanner.Value;
StartScope.WritePartialStatusMsg( outputline, FunctionName );
Scanner.GetToken;
// this call the ActivateScope
ActivateScope( FunctionName, true );
try try
  CurrentIdentScope.ExitLabel := 'Exit_'+CurrentIdentScope.ScopeName+ '_'+ GetLabel;
  // parse the param list
  LocalVarList := '';
  Scanner.Match( CSTI_OpenRound );
  While Scanner.Token = CSTI_Identifier do
    begin
    NewIdent( Scanner.Value,IsLocalVar );
    Scanner.GetToken;
    if not Scanner.IfMatch( CSTI_Comma ) then
      break
    end;
  Scanner.Match( CSTI_CloseRound );
  StartScope.WritePartialStatusMsg( outputline, '( ' +Copy( LocalVarList, 1, Length(LocalVarList)-2)+ ' )' );
  // do the actual function body
  DoStatementBlock( true );

  if (AnsiPos('ret',CurrentIdentScope.CodeSnipet[CurrentIdentScope.CodeSnipet.count-1]) <= 0) and
     (AnsiPos('ret',CurrentIdentScope.CodeSnipet[CurrentIdentScope.CodeSnipet.count-2]) <= 0) and
     AutoEmitLastRet then
    begin
    DoPushConst( 0 );
    DoReturn( False );
    end;
  CurrentIdentScope.WriteMsg( CurrentIdentScope.ExitLabel + ':' );

finally
  DeactivateCurrentScope;
end;
except
  On E : Exception do
  Raise ECompileError.Create( 'In function "' + FunctionName + '" ' + E.message ) at ExceptAddr;
end;
end; {DoInlinableFunction}

Procedure TBOSCompiler.DoFunction;
var
  FunctionName : string;
  index,StartProcLineNumber : integer;
begin
if Scanner.Token <> CSTI_Identifier then
  Scanner.Match( CSTI_Identifier );
FunctionName := Scanner.Value;
Scanner.GetToken;
ActivateScope( FunctionName, false );
try try
  CurrentIdentScope.WriteMsg( '' );
  CurrentIdentScope.WriteMsg( 'proc ' + FunctionName);
  // parse the param list
  LocalVarList := '';
  Scanner.Match( CSTI_OpenRound );
  While Scanner.Token = CSTI_Identifier do
    begin
    NewIdent( Scanner.Value,IsLocalVar );
    Scanner.GetToken;
    if not Scanner.IfMatch( CSTI_Comma ) then
      break
    end;
  Scanner.Match( CSTI_CloseRound );
  index := 1+CurrentIdentScope.WriteMsg( 'args none;');
  StartProcLineNumber := CurrentIdentScope.WriteMsg( 'beginproc' );
  // do the actual function body
  LoopCount := 0;
  DoStatementBlock( true );
  // check to see if the proc is empty
  if (CurrentIdentScope.CodeSnipet.Count - StartProcLineNumber <= 3) then
    begin
    CurrentIdentScope.CodeSnipet.Delete(CurrentIdentScope.CodeSnipet.Count-1);
    CurrentIdentScope.CodeSnipet.Delete(CurrentIdentScope.CodeSnipet.Count-1);
    CurrentIdentScope.WriteMsg( 'empty;' );  
    end;
  if LocalVarList <> '' then
    setlength( LocalVarList, length( LocalVarList ) -2);
  if LocalVarList = '' then
    LocalVarList := 'none';
  CurrentIdentScope.WriteMsg( index, 'var ' + LocalVarList+';');
  if (AnsiPos('ret',CurrentIdentScope.CodeSnipet[CurrentIdentScope.CodeSnipet.count-1]) <= 0) and
     (AnsiPos('ret',CurrentIdentScope.CodeSnipet[CurrentIdentScope.CodeSnipet.count-2]) <= 0) and
     AutoEmitLastRet then
    begin
    DoPushConst( 0 );
    DoReturn( False );
    end;
  CurrentIdentScope.WriteMsg( 'endproc;' );
  CurrentIdentScope.WriteMsg( '' );
finally
  Assert( CurrentIdentScope <> nil);
  CurrentIdentScope.CodeSnipet.WriteTo( CurrentIdentScope.fParentScope );
  DeactivateCurrentScope;
end;
except
  On E : Exception do
  Raise ECompileError.Create( 'In function "' + FunctionName + '" ' + E.message ) at ExceptAddr;
end;
end; {DoFunction}

Procedure TBOSCompiler.DoStatementBlock( ForceBlock : boolean = false );
begin
inc( CurrentIdentScope.IndentLevel, 2 );
try
  If ForceBlock or Scanner.IfMatch( CSTII_Begin ) then
    begin
    if ForceBlock then
      Scanner.Match( CSTII_Begin );
    CurrentIdentScope.WriteMsg( '// {' );
    while DoStatement do ;
    Scanner.match( CSTII_End );
    CurrentIdentScope.WriteMsg( '// }' );
    end
  else
    DoStatement;
finally
  dec( CurrentIdentScope.IndentLevel, 2 );
end;
end; {DoStatementBlock}

function TBOSCompiler.DoStatement : boolean;
begin
result := true;
case Scanner.Token of
  CSTII_LocalVar : DoLocalVar;
  CSTII_const : DoConst;
  CSTII_if : DoIf;
  CSTII_While : DoWhile;
  CSTII_do : DoDoWhile;
  CSTII_for : DoFor;
  CSTII_Break : DoBreak;
  CSTII_Continue : DoContinue;
  CSTII_Return : DoReturn( True );
  CSTI_SemiColon : Scanner.Match( CSTI_SemiColon ); // null statement
  CSTI_Identifier: DoAssignment;
  CSTII_BASMInline : DoBASMInline;
//  CSTII_GetUnitProperty : DoGetUnitProperty;
//  CSTII_Rand : DoRand;
  CSTII_SetUnitProperty : DoSetUnitProperty;
  CSTII_Sleep : DoSleep;
  CSTII_Turn : DoTurn;
  CSTII_WaitForTurn : DoWaitForTurn;
  CSTII_Move : DoMove;
  CSTII_WaitForMove : DoWaitForMove;
  CSTII_Spin :  DoSpin;
  CSTII_StopSpin : DoStopSpin;
  CSTII_Hide : DoHide;
  CSTII_Show : DoShow;
  CSTII_Cache : DoCache;
  CSTII_DontCache : DoDontCache;
  CSTII_CallScript : DoCallScript( false, 0, False );
  CSTII_StartScript : DoStartScript;
  CSTII_Signal : DoSignal;
  CSTII_SetSignalMask : DoSetSignalMask;
  CSTII_Explode : DoExplode;
  CSTII_EmitSfx : DoEmitSfx;
  CSTII_AttachUnit : DoAttachUnit;
  CSTII_DropUnit : DoDropUnit;
  CSTII_Shade : DoShade;
  CSTII_DontShade : DoDontShade;
  CSTI_Increment : DoPreIncrement;
  CSTI_Decrement : DoPreDecrement;
else result := false;
end; {case}
end; {DoStatement}

Procedure TBOSCompiler.DoLocalVar;
var
  LocalVarIdent : TIdent;
  OutputLine : integer;
begin
Scanner.Match( CSTII_LocalVar );
repeat
  if Scanner.Token <> CSTI_Identifier then
    Scanner.Match( CSTI_Identifier );
  LocalVarIdent := NewIdent( Scanner.Value, IsLocalVar );
  Scanner.GetToken;
  if (Scanner.Token = CSTI_Assignment) then
    begin
    Scanner.GetToken;
    CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'MakeLocalVar '+LocalVarIdent.Name+' = ( ');
    DoExpression( OutputLine );
    CurrentIdentScope.WritePartialStatusMsg( OutputLine, ');');
    DoPop( LocalVarIdent );
    end;
until not Scanner.IfMatch( CSTI_Comma );
Scanner.Match( CSTI_SemiColon );
{
setlength( VarList, length( VarList )-2 );
CurrentIdentScope.WriteMsg( 'MakeLocalVar ' + VarList );
}
end; {DoLocalVar}

Procedure TBOSCompiler.DoIf;
var
  Label1 : string;
  Label2 : string;
  OutputLine : integer;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'if ( ');
Label1 := GetLabel;
Scanner.Match( CSTII_if );
Scanner.Match( CSTI_OpenRound );
DoExpression( OutputLine );
//CurrentIdentScope.WriteMsg( 'neg;' );
CurrentIdentScope.WriteMsg( 'jnz iflabel_' + Label1 + ';' );
Scanner.Match( CSTI_CloseRound );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, ' )');
DoStatementBlock;
If Scanner.IfMatch( CSTII_Else ) then
  begin
  CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'else');
  Label2 := GetLabel;
  CurrentIdentScope.WriteMsg( 'jmp if_end_label_' + Label2 + ';' );
  CurrentIdentScope.WriteMsg( 'iflabel_' + Label1 + ':' );
  DoStatementBlock;
  CurrentIdentScope.WriteMsg( 'if_end_label_' + Label2 + ':' );
  end
else
  CurrentIdentScope.WriteMsg( 'iflabel_' + Label1 + ':' );
CurrentIdentScope.WriteMsg( '' );
end; {DoIf}

Procedure TBOSCompiler.DoWhile;
var
  Label1 : string;
  Label2 : string;
  OldLoopContinueLabel, OldLoopBreakLabel : string;
  OldUsedBreak : boolean;
  OutputLine : integer;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'While (');
OldLoopContinueLabel := LoopContinueLabel;
OldLoopBreakLabel := LoopBreakLabel;
OldUsedBreak := UsedBreak;
inc( LoopCount );
try
  Label1 := GetLabel;
  Label2 := GetLabel;
  LoopContinueLabel := 'ConditionLabel_' + Label1;
  LoopBreakLabel := 'TerminatorLabel_' + Label2;
  CurrentIdentScope.WriteMsg( LoopContinueLabel + ':' );
  CurrentIdentScope.WritePartialStatusMsg( OutputLine, ' ) ');
  Scanner.Match( CSTII_while );
  Scanner.Match( CSTI_OpenRound );
  DoExpression( OutputLine );
  CurrentIdentScope.WriteMsg( 'jnz ' + LoopBreakLabel + ';' );
  Scanner.Match( CSTI_CloseRound );
  DoStatementBlock;
  CurrentIdentScope.WriteMsg( 'jmp ' + LoopContinueLabel + ';' );
  CurrentIdentScope.WriteMsg( LoopBreakLabel + ':' );
  CurrentIdentScope.WriteMsg( '' );
finally
  LoopContinueLabel := OldLoopContinueLabel;
  LoopBreakLabel := OldLoopBreakLabel;
  UsedBreak := OldUsedBreak;
  dec( LoopCount );
end;
end; {DoWhile}

Procedure TBOSCompiler.DoDoWhile;
var
  Label1 : string;
  Label2 : string;
  OldLoopContinueLabel, OldLoopBreakLabel : string;
  OldUsedBreak : boolean;
  OutputLine : integer;
begin
OldLoopContinueLabel := LoopContinueLabel;
OldLoopBreakLabel := LoopBreakLabel;
OldUsedBreak := UsedBreak;
inc( LoopCount );
try
  Label1 := GetLabel;
  Label2 := GetLabel;
  LoopContinueLabel := 'ConditionLabel_' + Label1;
  LoopBreakLabel := 'TerminatorLabel_' + Label2;
  CurrentIdentScope.WriteMsg( LoopContinueLabel + ':' );
  CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'Do');
  Scanner.Match( CSTII_Do );
  DoStatementBlock;
  CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'while (');
  Scanner.Match( CSTII_while );
  Scanner.Match( CSTI_OpenRound );
  DoExpression( OutputLine );
  CurrentIdentScope.WriteMsg( 'neg;' );
  CurrentIdentScope.WriteMsg( 'jnz ' + LoopContinueLabel + ';' );
  CurrentIdentScope.WritePartialStatusMsg( OutputLine, ' ) ');
  Scanner.Match( CSTI_CloseRound );
  if UsedBreak then
    CurrentIdentScope.WriteMsg( LoopBreakLabel + ':' );
  CurrentIdentScope.WriteMsg( '' );
finally
  LoopContinueLabel := OldLoopContinueLabel;
  LoopBreakLabel := OldLoopBreakLabel;
  UsedBreak := OldUsedBreak;
  dec( LoopCount );
end;
end; {DoDoWhile}

Procedure TBOSCompiler.DoFor;
var
  Label1 : string;
  Label2 : string;
  OldLoopContinueLabel, OldLoopBreakLabel : string;
  OldUsedBreak : boolean;
   OutputLine : integer;
begin
OldLoopContinueLabel := LoopContinueLabel;
OldLoopBreakLabel := LoopBreakLabel;
OldUsedBreak := UsedBreak;
inc( LoopCount );
try
  Label1 := GetLabel;
  Label2 := GetLabel;
  LoopContinueLabel := 'ConditionLabel_' + Label1;
  LoopBreakLabel := 'TerminatorLabel_' + Label2;
  CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'for ( ');
  Scanner.Match( CSTII_for );
  Scanner.Match( CSTI_OpenRound );
  // start expression
  repeat
    DoExpression( OutputLine );
    if scanner.token = CSTI_Comma then
      CurrentIdentScope.WritePartialStatusMsg( OutputLine, ', ');
  until not Scanner.IfMatch( CSTI_Comma );
  CurrentIdentScope.WritePartialStatusMsg( OutputLine, ' ; ');
  Scanner.Match( CSTI_SemiColon );
  CurrentIdentScope.WriteMsg( LoopContinueLabel + ':' );
  repeat
    DoExpression( OutputLine );
    if scanner.token = CSTI_Comma then
      CurrentIdentScope.WritePartialStatusMsg( OutputLine, ', ');
  until not Scanner.IfMatch( CSTI_Comma );
  CurrentIdentScope.WriteMsg( 'jnz ' + LoopBreakLabel + ';' );
  CurrentIdentScope.WritePartialStatusMsg( OutputLine, ' ; ');
  Scanner.Match( CSTI_SemiColon );
  repeat
    DoExpression( OutputLine );
    if scanner.token = CSTI_Comma then
      CurrentIdentScope.WritePartialStatusMsg( OutputLine, ', ');
  until not Scanner.IfMatch( CSTI_Comma );
  CurrentIdentScope.WritePartialStatusMsg( OutputLine, ' ) ');
  Scanner.Match( CSTI_CloseRound );
  DoStatementBlock;
  CurrentIdentScope.WriteMsg( 'jmp ' + LoopContinueLabel + ';' );
  CurrentIdentScope.WriteMsg( LoopBreakLabel + ':' );
  CurrentIdentScope.WriteMsg( '' );
finally
  LoopContinueLabel := OldLoopContinueLabel;
  LoopBreakLabel := OldLoopBreakLabel;
  UsedBreak := OldUsedBreak;  
  dec( LoopCount );
end;
end; {DoFor}

Procedure TBOSCompiler.DoBreak;
var
  OutputLine : integer;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'Break' );
Scanner.Match( CSTII_Break );
//WriteDWord( 
//CurrentIdentScope.WriteMsg( 'break' );
CurrentIdentScope.WriteMsg( 'jmp ' + LoopBreakLabel + ';' );
UsedBreak := true;
end; {DoBreak}

Procedure TBOSCompiler.DoContinue;
var
  OutputLine : integer;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'Continue' );
Scanner.Match( CSTII_Continue );
CurrentIdentScope.WriteMsg( 'jmp ' + LoopContinueLabel + ';' );
end; {DoContinue}

Procedure TBOSCompiler.DoReturn( UseMarker : Boolean );
var
  OutputLine : integer;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'return ' );
if UseMarker then
  begin
  Scanner.Match( CSTII_Return );
  DoExpression( OutputLine );
  end;
if not CurrentIdentScope.Inlineable then
  CurrentIdentScope.WriteMsg( 'ret;' )
else
  CurrentIdentScope.WriteMsg( 'jmp '+CurrentIdentScope.ExitLabel + ';' );
end; {DoReturn}

Procedure TBOSCompiler.DoGetUnitProperty( OutputLine : integer );
begin
CurrentIdentScope.WritePartialStatusMsg( OutputLine, 'Get ' );
Scanner.Match( CSTII_GetUnitProperty );
DoExpression( OutputLine );
if ( Scanner.Token = CSTI_OpenRound ) then
  begin
  BiuldParamList( OutputLine );
  CurrentIdentScope.WriteMsg( 'get;' );
  end
else
  CurrentIdentScope.WriteMsg( 'getuv;' );
//WriteInstruction( Opcode_GetUnitValue );
// TODO 3 : TBOSCompiler.DoGetUnitProperty finish properly
end; {DoGetUnitProperty}

Procedure TBOSCompiler.DoSetUnitProperty;
var
  OutputLine : integer;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'Set ' );
Scanner.Match( CSTII_SetUnitProperty );
DoExpression( OutputLine );
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'to ' );
Scanner.Match( CSTII_To );
DoExpression( OutputLine );
CurrentIdentScope.WriteMsg( 'set;' );
end; {DoSetUnitProperty}

Procedure TBOSCompiler.DoSleep;
var
  OutputLine : integer;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'sleep ' );
Scanner.Match( CSTII_Sleep ); 
DoExpression( OutputLine );
CurrentIdentScope.WriteMsg( 'sleep;' );
end; {DoSleep}

Procedure TBOSCompiler.DoTurn;
var
  aPieceName, aAxisName : string;
  OutputLine : integer;
  CodeSnipetID1, CodeSnipetID2 : TCodeSnipet;
  TurnWithSpeed : Boolean;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'turn ' );
Scanner.Match( CSTII_Turn );
ValidatePiece( aPieceName );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, aPieceName+' to ' );
Scanner.Match( CSTII_To );
ValidateAxis( aAxisName );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, aAxisName+' ' );
CodeSnipetID1 := nil; CodeSnipetID2 := nil;
try
  CodeSnipetID1 := TCodeSnipet.Create( CurrentIdentScope );
  CodeSnipetID2 := TCodeSnipet.Create( CurrentIdentScope );
  // the heading
  CodeSnipetID1.Hook( CurrentIdentScope.CodeSnipet.Count );
  DoExpression( OutputLine );
  TurnWithSpeed := scanner.IfMatch( CSTII_Speed );
  if TurnWithSpeed then
    begin
    CurrentIdentScope.WritePartialStatusMsg( OutputLine, 'speed ' );
    CodeSnipetID2.Hook( OutputLine+1 );
    try
      DoExpression( OutputLine );
    finally
      CodeSnipetID2.UnHook;
      CodeSnipetID1.UnHook;
    end;
    end;
finally
  try
    if (CodeSnipetID2 <> nil)  then
      begin
      if CodeSnipetID2.fCapturing then
        CodeSnipetID2.UnHook;
      CodeSnipetID2.Write;
      end;
    if (CodeSnipetID1 <> nil) then
      begin
      if CodeSnipetID1.fCapturing then
        CodeSnipetID1.UnHook;
      CodeSnipetID1.Write;
      end;
  finally
    FreeAndNil( CodeSnipetID2 );
    FreeAndNil( CodeSnipetID1 );
  end;
end;
if TurnWithSpeed  then
  begin
  CurrentIdentScope.WriteMsg( 'turn;');
  end
else
  begin
  scanner.Match( CSTII_Now );
  CurrentIdentScope.WritePartialStatusMsg( OutputLine, 'now ' );
  CurrentIdentScope.WriteMsg( 'turnnow;');
  end;
CurrentIdentScope.WriteMsg( 'ddw ' + aPieceName + ';' );
CurrentIdentScope.WriteMsg( 'ddw ' + aAxisName + ';' );
end; {DoTurn}

Procedure TBOSCompiler.DoWaitForTurn;
var
  aPieceName, aAxisName : string;
  OutputLine : integer;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'wait-for-turn ' );
Scanner.Match( CSTII_WaitForTurn );
ValidatePiece( aPieceName );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, aPieceName+' around ' );
Scanner.Match( CSTII_around );
ValidateAxis( aAxisName );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, aAxisName+' ' );
CurrentIdentScope.WriteMsg( 'wait-for-turn;');
CurrentIdentScope.WriteMsg( 'ddw ' + aPieceName + ';' );
CurrentIdentScope.WriteMsg( 'ddw ' + aAxisName + ';' );
end; {DoWaitForTurn}

Procedure TBOSCompiler.DoMove;
var
  aPieceName, aAxisName : string;
  OutputLine : integer;
  CodeSnipetID1, CodeSnipetID2 : TCodeSnipet;
  MoveWithSpeed : Boolean;

begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'move ' );
Scanner.Match( CSTII_Move );
ValidatePiece( aPieceName );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, aPieceName+' to ' );
Scanner.Match( CSTII_To );
ValidateAxis( aAxisName );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, aAxisName+' ' );
CodeSnipetID1 := nil; CodeSnipetID2 := nil;
try
  CodeSnipetID1 := TCodeSnipet.Create( CurrentIdentScope );
  CodeSnipetID2 := TCodeSnipet.Create( CurrentIdentScope );
  // the heading
  CodeSnipetID1.Hook( CurrentIdentScope.CodeSnipet.Count );
  DoExpression( OutputLine );
  MoveWithSpeed := scanner.IfMatch( CSTII_Speed );
  if MoveWithSpeed then
    begin
    CurrentIdentScope.WritePartialStatusMsg( OutputLine, 'speed ' );

    CodeSnipetID2.Hook( OutputLine+1 );
    try
      DoExpression( OutputLine );
    finally
      CodeSnipetID2.UnHook;
      CodeSnipetID1.UnHook;
    end;
    end;
finally
  try
    if (CodeSnipetID2 <> nil)  then
      begin
      if CodeSnipetID2.fCapturing then
        CodeSnipetID2.UnHook;
      CodeSnipetID2.Write;
      end;
    if (CodeSnipetID1 <> nil) then
      begin
      if CodeSnipetID1.fCapturing then
        CodeSnipetID1.UnHook;
      CodeSnipetID1.Write;
      end;
  finally
    FreeAndNil( CodeSnipetID2 );
    FreeAndNil( CodeSnipetID1 );
  end;
end;
if MoveWithSpeed then
  CurrentIdentScope.WriteMsg( 'move;' )
else
  begin
  scanner.Match( CSTII_Now );
  CurrentIdentScope.WritePartialStatusMsg( OutputLine, 'Now ' );
  CurrentIdentScope.WriteMsg( 'movenow;' );
  end;
CurrentIdentScope.WriteMsg( 'ddw ' + aPieceName + ';' );
CurrentIdentScope.WriteMsg( 'ddw ' + aAxisName + ';' );
end; {DoMove}

Procedure TBOSCompiler.DoWaitForMove;
var
  aPieceName, aAxisName : string;
  OutputLine : integer;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'wait-for-move ' );
Scanner.Match( CSTII_WaitForMove );
ValidatePiece( aPieceName );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, aPieceName+' along ' );
Scanner.Match( CSTII_along );
ValidateAxis( aAxisName );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, aAxisName+' ' );
CurrentIdentScope.WriteMsg( 'wait-for-move;');
CurrentIdentScope.WriteMsg( 'ddw ' + aPieceName + ';' );
CurrentIdentScope.WriteMsg( 'ddw ' + aAxisName + ';' );
end; {DoWaitForMove}

Procedure TBOSCompiler.DoSpin;
var
  aPieceName, aAxisName : string;
  OutputLine : integer;
begin { TODO : lookup full syntax for spin }
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'spin ' );
Scanner.Match( CSTII_Spin );
ValidatePiece( aPieceName );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, aPieceName+' around ' );
Scanner.Match( CSTII_around );
ValidateAxis( aAxisName );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, aAxisName+' ' );
DoExpression( OutputLine );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, 'speed ' );
scanner.Match( CSTII_Speed );
DoExpression( OutputLine );
If scanner.IfMatch( CSTII_Accelerate ) then
  begin
  CurrentIdentScope.WritePartialStatusMsg( OutputLine, 'accelerate ' );
  DoExpression( OutputLine );
  end
else
  DoPushConst( 0 );
CurrentIdentScope.WriteMsg( 'spin;');
CurrentIdentScope.WriteMsg( 'ddw ' + aPieceName + ';' );
CurrentIdentScope.WriteMsg( 'ddw ' + aAxisName + ';' );
end; {DoSpin}

Procedure TBOSCompiler.DoStopSpin;
var
  aPieceName, aAxisName : string;
  OutputLine : integer;
begin { TODO : lookup full syntax for stop-spin }
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'stop-spin ' );
Scanner.Match( CSTII_StopSpin );
ValidatePiece( aPieceName );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, aPieceName+' around ' );
Scanner.Match( CSTII_around );
ValidateAxis( aAxisName );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, aAxisName+' ' );
DoExpression( OutputLine );
If scanner.IfMatch( CSTII_Decelerate ) then
  begin
  CurrentIdentScope.WritePartialStatusMsg( OutputLine, 'decelerate ' );
  DoExpression( OutputLine );
  end
else
  DoPushConst( 0 );
CurrentIdentScope.WriteMsg( 'stop-spin;');
CurrentIdentScope.WriteMsg( 'ddw ' + aPieceName + ';' );
CurrentIdentScope.WriteMsg( 'ddw ' + aAxisName + ';' );
end; {DoStopSpin}

Procedure TBOSCompiler.DoHide;
var
  aPieceName : string;
  OutputLine : integer;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'hide ' );
Scanner.Match( CSTII_Hide );
ValidatePiece( aPieceName );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, aPieceName+' ' );
CurrentIdentScope.WriteMsg( 'hide;');
CurrentIdentScope.WriteMsg( 'ddw ' + aPieceName + ';' );
end; {DoHide}

Procedure TBOSCompiler.DoShow;
var
  aPieceName : string;
  OutputLine : integer;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'show ' );
Scanner.Match( CSTII_Show );
ValidatePiece( aPieceName );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, aPieceName+' ' );
CurrentIdentScope.WriteMsg( 'show;');
CurrentIdentScope.WriteMsg( 'ddw ' + aPieceName + ';' );
end; {DoShow}

Procedure TBOSCompiler.DoCache;
var
  aPieceName : string;
  OutputLine : integer;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'cache  ' );
Scanner.Match( CSTII_Cache );
ValidatePiece( aPieceName );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, aPieceName+' ' );
CurrentIdentScope.WriteMsg( 'cache;');
CurrentIdentScope.WriteMsg( 'ddw ' + aPieceName + ';' );
end; {DoCache}

Procedure TBOSCompiler.DoDontCache;
var
  aPieceName : string;
  OutputLine : integer;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'dont-cache  ' );
Scanner.Match( CSTII_DontCache );
ValidatePiece( aPieceName );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, aPieceName+' ' );
CurrentIdentScope.WriteMsg( 'dont-cache;');
CurrentIdentScope.WriteMsg( 'ddw ' + aPieceName + ';' );
end; {DoDontCache}

Procedure TBOSCompiler.DoShade;
var
  aPieceName : string;
  OutputLine : integer;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'shade  ' );
Scanner.Match( CSTII_Shade );
ValidatePiece( aPieceName );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, aPieceName+' ' );
CurrentIdentScope.WriteMsg( 'shade;');
CurrentIdentScope.WriteMsg( 'ddw ' + aPieceName + ';' );
end; {DoShade}

Procedure TBOSCompiler.DoDontShade;
var
  aPieceName : string;
  OutputLine : integer;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'dont-shade  ' );
Scanner.Match( CSTII_DontShade );
ValidatePiece( aPieceName );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, aPieceName+' ' );
CurrentIdentScope.WriteMsg( 'dont-shade;');
CurrentIdentScope.WriteMsg( 'ddw ' + aPieceName + ';' );
end; {DoDontShade}

Function TBOSCompiler.BiuldParamList( OutputLine : integer ) : integer;
begin
result := 0;
CurrentIdentScope.WritePartialStatusMsg( OutputLine, '( ');
Scanner.Match( CSTI_OpenRound );
While not Scanner.IfMatch( CSTI_CloseRound ) do
  begin
  DoExpression( OutputLine );
  inc( result );
  if not Scanner.IfMatch( CSTI_Comma ) then
    begin
    Scanner.Match( CSTI_CloseRound );
    break;
    end;
  CurrentIdentScope.WritePartialStatusMsg( OutputLine, ', ');
  end;
CurrentIdentScope.WritePartialStatusMsg( OutputLine, ' )');
end; {BiuldParamList}

Procedure TBOSCompiler.DoCallScript( CanReturnValue : boolean{ = false};
                                     OutputLine : integer{ = 0};
                                     NoCallScriptMarker : boolean{ = False});
var
  FunctionName : string;
  Index,ParamCount : integer;
  Ident : TIdentScope;
  b : Boolean;
  tmp : TObject;
begin
if CanReturnValue then
  CurrentIdentScope.WritePartialStatusMsg( OutputLine, 'call-script ' )
else
  CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'call-script ' );
if not NoCallScriptMarker then
  Scanner.Match( CSTII_CallScript );
FunctionName := GetFunc;
CurrentIdentScope.WritePartialStatusMsg( OutputLine, FunctionName+' ' );

Ident := CurrentIdentScope.FindSubScope( FunctionName, b );
if (Ident <> nil ) and not b and Ident.Inlineable then
  begin
  // check for cyclic inbedding
  if Ident.IsParent( CurrentIdentScope ) then
    WriteErrorMsg( FatalError, 'Both '+FunctionName+' and '+
                               CurrentIdentScope.ScopeName+
                               ' use each other. Cyclic inclusion of inlinable functions not allowed' );
  // inject the inlinable local variables & consts into the function (gets a list of them too)
  for Index := 0 to Ident.fIdents.count-1 do
    begin
    tmp := TObject( Ident.fIdents.objects[ Index ] );
    if (tmp is TIdent) and
      (TIdent(tmp).flags in [ IsLocalVar,IsConst ]) and
      (( FindIdent( TIdent(tmp).Name, b ) = nil) or b) then
      NewIdent( TIdent(tmp).Name, TIdent(tmp).flags );
    end;
  // biuld the param list & assign the local variables
  ParamCount := 0;
  CurrentIdentScope.WritePartialStatusMsg( OutputLine, '( ');
  Scanner.Match( CSTI_OpenRound );
  While not Scanner.IfMatch( CSTI_CloseRound ) do
    begin
    // get the expression
    DoExpression( OutputLine );
    if ParamCount < Ident.LocalVarCount then
      begin
      // find the variable to set
      DoPop( Ident.LocalVar[ ParamCount ] );
      inc( ParamCount );
      end;
    if not Scanner.IfMatch( CSTI_Comma ) then
      begin
      Scanner.Match( CSTI_CloseRound );
      break;
      end;
    CurrentIdentScope.WritePartialStatusMsg( OutputLine, ', ');
    end;
  if ParamCount > Ident.LocalVarCount then
    WriteErrorMsg( NormalError, 'Too many parameters, '+IntToStr( ParamCount )+
                                ', to inlined function '+FunctionName+
                                ', Max:'+IntToStr( Ident.LocalVarCount ) );
  CurrentIdentScope.WritePartialStatusMsg( OutputLine, ' )');
  Ident.CodeSnipet.WriteTo( CurrentIdentScope );
  end
else
  begin
  ParamCount := BiuldParamList( OutputLine );
  //WriteInstruction( Opcode_CallScript );
  //WriteInstrArg( FuncIndex );
  //WriteInstrArg( ParamCount );
  CurrentIdentScope.WriteMsg( 'call;');
  CurrentIdentScope.WriteMsg( 'ddw ' + FunctionName + ';' );
  CurrentIdentScope.WriteMsg( 'ddw ' + IntToStr( ParamCount ) + ';' );
  end;
// todo : Determine if "Call-Script"  does return anything
//if not CanReturnValue then
//  CurrentIdentScope.WriteMsg( 'flush;' );
end; {DoCallScript}

Procedure TBOSCompiler.DoStartScript;
var
  FunctionName : string;
  ParamCount : integer;
  OutputLine : integer;
  Ident : TIdentScope;
  ScopeNameIsIdent : Boolean;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'start-script ' );
Scanner.Match( CSTII_StartScript );
FunctionName := GetFunc;
CurrentIdentScope.WritePartialStatusMsg( OutputLine, FunctionName+' ' );
Ident := CurrentIdentScope.FindSubScope( FunctionName, ScopeNameIsIdent );
if (Ident <> nil ) and not ScopeNameIsIdent and Ident.Inlineable then
  WriteErrorMsg( FatalError, 'Inlinable function, '+FunctionName+', can not be called via start-script' );
ParamCount := BiuldParamList( OutputLine );
//WriteInstruction( Opcode_StartScript );
//WriteInstrArg( FuncIndex );
//WriteInstrArg( ParamCount );
CurrentIdentScope.WriteMsg( 'start;');
CurrentIdentScope.WriteMsg( 'ddw ' + FunctionName + ';' );
CurrentIdentScope.WriteMsg( 'ddw ' + IntToStr( ParamCount ) + ';' );
// todo : Determine if "Start-Script"  does return anything
//CurrentIdentScope.WriteMsg( 'flush;' );
end; {DoStartScript}

Procedure TBOSCompiler.DoSignal;
var
  OutputLine : integer;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'signal ' );
Scanner.Match( CSTII_Signal );
DoExpression( OutputLine );
CurrentIdentScope.WriteMsg( 'sig;' );
end; {DoSignal}

Procedure TBOSCompiler.DoSetSignalMask;
var
  OutputLine : integer;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'set-signal-mask ' );
Scanner.Match( CSTII_SetSignalMask );
DoExpression( OutputLine );
CurrentIdentScope.WriteMsg( 'mask;' );
end; {DoSetSignalMask}

Procedure TBOSCompiler.DoExplode;
var
  aPieceName : string;
  OutputLine : integer;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'explode ' );
Scanner.Match( CSTII_Explode );
ValidatePiece( aPieceName );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, aPieceName+' type ');
Scanner.Match( CSTII_ExplodeType );
DoExpression( OutputLine );
CurrentIdentScope.WriteMsg( 'explode;' );
CurrentIdentScope.WriteMsg( 'ddw ' + aPieceName + ';');
end; {DoExplode}

Procedure TBOSCompiler.DoEmitSfx;
var
  aPieceName : string;
  OutputLine : integer;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'emit-sfx ' );
Scanner.Match( CSTII_EmitSfx );
DoExpression( OutputLine );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, 'from ');
Scanner.Match( CSTII_from );
ValidatePiece( aPieceName );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, aPieceName+' ');
CurrentIdentScope.WriteMsg( 'sfx; ' );
CurrentIdentScope.WriteMsg( 'ddw ' + aPieceName + ';');
end; {DoEmitSfx}

Procedure TBOSCompiler.DoRand( OutputLine : integer );
begin
CurrentIdentScope.WritePartialStatusMsg( OutputLine, 'rand( ' );
Scanner.Match( CSTII_Rand );
Scanner.Match( CSTI_OpenRound );
DoExpression( OutputLine );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, ', ');
Scanner.Match( CSTI_Comma );
DoExpression( OutputLine );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, ') ');
Scanner.Match( CSTI_CloseRound );
CurrentIdentScope.WriteMsg( 'rand;' );
end; {DoRand}

Procedure TBOSCompiler.DoAttachUnit;
var
  OutputLine : integer;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'attach-unit ' );
Scanner.Match( CSTII_AttachUnit );
DoExpression( OutputLine );
CurrentIdentScope.WritePartialStatusMsg( OutputLine, 'to ' );
if Scanner.ifMatch( CSTII_to ) then
  DoExpression( OutputLine )
else
  DoPushConst( -1 );
DoPushConst( 0 );
CurrentIdentScope.WriteMsg( 'attach;' );
end; {DoAttachUnit}

Procedure TBOSCompiler.DoDropUnit;
var
  OutputLine : integer;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, 'drop-unit ' );
Scanner.Match( CSTII_DropUnit );
DoExpression( OutputLine );
CurrentIdentScope.WriteMsg( 'drop;' );
end; {DoDropUnit}

Procedure TBOSCompiler.DoBASMInline;
var
  OutputLine : integer;
  BASMLine : string;
begin // DONE 2 : Allow inlining BASM code
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, '__basm' );
Scanner.Match( CSTII_BASMInline );
inc( CurrentIdentScope.IndentLevel, 2 );
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, '{' );
try
  Scanner.Match( CSTII_Begin );
  While not Scanner.IfMatch( CSTII_end ) do
    begin
    BASMLine := '';
    While not Scanner.IfMatch( CSTI_SemiColon ) and
          not Scanner.IfMatch( CSTII_end ) do
      begin
      BASMLine := BASMLine +Scanner.Value + ' ';
      Scanner.GetToken;
      end;
    Delete(BASMLine, length(BASMLine),1);
    BASMLine := BASMLine +';';
    CurrentIdentScope.WriteMsg( BASMLine );
    end;
finally
  CurrentIdentScope.WriteBeginStatusMsg( OutputLine, '}' );
  dec( CurrentIdentScope.IndentLevel, 2 );
end;
end; {DoBASMInline}

Procedure TBOSCompiler.AssignmentHelper( OutputLine : integer;
                                         var ResultIdent : TIdent);
var
  ResultName : string;
  a : boolean;                                           
begin
if Scanner.Token <> CSTI_Identifier then
  Scanner.Match( CSTI_Identifier );
ResultName := Scanner.Value;
CurrentIdentScope.WritePartialStatusMsg( OutputLine, ResultName +' ');
Scanner.GetToken;
ResultIdent := FindIdent( ResultName, a );
if ResultIdent = nil then
  begin
  WriteErrorMsg( NormalError, 'Undeclaired identifier, ' + ResultName );
  ResultIdent := NewIdent( ResultName, UnknownVarType );
  end;
end; {AssignmentHelper}

Procedure TBOSCompiler.DoAssignment;
var
  ResultIdent : TIdent;
  OutputLine : integer;
//  OutputOpcode : longword;
  AssignmentType : TToken;
  ValidAssignmentTokens : set of TToken;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, '');
AssignmentHelper( OutputLine, ResultIdent );
ValidAssignmentTokens := [ CSTI_Assignment_Plus, CSTI_Assignment_Minus,
                           CSTI_Assignment_Divide, CSTI_Assignment_Multiply,
                           CSTI_Assignment_BitwiseOr, CSTI_Assignment_BitwiseXor,
                           CSTI_Assignment_BitwiseAnd, CSTI_Assignment_LogicalOr,
                           CSTI_Assignment_LogicalXor, CSTI_Assignment_LogicalAnd ];
if EnableBASMExtentions then
  begin
  include( ValidAssignmentTokens, CSTI_Assignment_ShiftLeft );
  include( ValidAssignmentTokens, CSTI_Assignment_ShiftRight );
  end;
if Scanner.Token = CSTI_Assignment then
  begin
  CurrentIdentScope.WritePartialStatusMsg(OutputLine, '= ');
  Scanner.GetToken;
  DoExpression( OutputLine );
  end
else if Scanner.Token in ValidAssignmentTokens then
  begin
  AssignmentType := Scanner.Token;
  Scanner.GetToken;
  case AssignmentType of
    CSTI_Assignment_Plus       : CurrentIdentScope.WritePartialStatusMsg( '+= ' );
    CSTI_Assignment_Minus      : CurrentIdentScope.WritePartialStatusMsg( '-= ' );
    CSTI_Assignment_Divide     : CurrentIdentScope.WritePartialStatusMsg( '/= ' );
    CSTI_Assignment_Multiply   : CurrentIdentScope.WritePartialStatusMsg( '*= ' );
    CSTI_Assignment_BitwiseOr  : CurrentIdentScope.WritePartialStatusMsg( '|= ' );
    CSTI_Assignment_BitwiseXor : CurrentIdentScope.WritePartialStatusMsg( '^= ' );
    CSTI_Assignment_BitwiseAnd : CurrentIdentScope.WritePartialStatusMsg( '&= ' );
    CSTI_Assignment_LogicalOr  : CurrentIdentScope.WritePartialStatusMsg( '||= ' );
    CSTI_Assignment_LogicalXor : CurrentIdentScope.WritePartialStatusMsg( '^^= ' );
    CSTI_Assignment_LogicalAnd : CurrentIdentScope.WritePartialStatusMsg( '&&= ' );
    CSTI_Assignment_ShiftLeft  : CurrentIdentScope.WritePartialStatusMsg( '<<= ' );
    CSTI_Assignment_ShiftRight : CurrentIdentScope.WritePartialStatusMsg( '>>= ' );
  end; {case}
  DoPush( ResultIdent );
  DoExpression( OutputLine );
  case AssignmentType of
    CSTI_Assignment_Plus       : CurrentIdentScope.WriteMsg( 'add;' );
    CSTI_Assignment_Minus      : CurrentIdentScope.WriteMsg( 'sub;' );
    CSTI_Assignment_Divide     : CurrentIdentScope.WriteMsg( 'div;' );
    CSTI_Assignment_Multiply   : CurrentIdentScope.WriteMsg( 'mul;' );
    CSTI_Assignment_BitwiseOr  : CurrentIdentScope.WriteMsg( 'or;' );
    CSTI_Assignment_BitwiseXor : CurrentIdentScope.WriteMsg( 'xor;' );
    CSTI_Assignment_BitwiseAnd : CurrentIdentScope.WriteMsg( 'and;' );
    CSTI_Assignment_LogicalOr  : CurrentIdentScope.WriteMsg( 'lor;' );
    CSTI_Assignment_LogicalXor : CurrentIdentScope.WriteMsg( 'lxor;' );
    CSTI_Assignment_LogicalAnd : CurrentIdentScope.WriteMsg( 'land;' );
    CSTI_Assignment_ShiftLeft  : CurrentIdentScope.WriteMsg( 'shl;' );
    CSTI_Assignment_ShiftRight : CurrentIdentScope.WriteMsg( 'sar;' );
  end; {case}
  end
else
  begin
  WriteErrorMsg( FatalError, 'Expecting assignment type' );
  Scanner.Match( CSTI_Assignment );
  end;
DoPop( ResultIdent );
//WriteInstruction( OutputOpcode );
//if OutputOpcode <> 0 then
//  WriteInstrArg( ResultIdent.IndexValue );
end; {DoAssignment}


Procedure TBOSCompiler.DoPreIncrement;
var
  OutputLine : longint;
  ResultIdent : TIdent;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, '++ ' );
Scanner.Match( CSTI_Increment );
AssignmentHelper( OutputLine, ResultIdent );
DoPush( ResultIdent );
DoPushConst( 1 );
CurrentIdentScope.WriteMsg( 'add;' );
DoPop( ResultIdent );
end; {DoPreIncrement}

Procedure TBOSCompiler.DoPreDecrement;
var
  OutputLine : longint;
  ResultIdent : TIdent;
begin
CurrentIdentScope.WriteBeginStatusMsg( OutputLine, '-- ' );
Scanner.Match( CSTI_Decrement );
AssignmentHelper( OutputLine, ResultIdent );
DoPush( ResultIdent );
DoPushConst( 1 );
CurrentIdentScope.WriteMsg( 'sub;' );
DoPop( ResultIdent );
end; {DoPreDecrement}

// Parse and Translate a Math Factor
Procedure TBOSCompiler.DoFactor( OutputLine : integer );
var
  Ident : TIdent;
  IntValue : Int64;
  IdentNameIsScope : boolean;
begin
case Scanner.token of
  CSTII_GetUnitProperty : DoGetUnitProperty( OutputLine );
  CSTII_Rand : DoRand( OutputLine );
  CSTII_CallScript : DoCallScript( true, OutputLine, false );
  CSTI_Increment :
    begin
    CurrentIdentScope.WritePartialStatusMsg( OutputLine, '++ ' );
    Scanner.Match( CSTI_Increment );
    AssignmentHelper( OutputLine, ident );
    DoPush( ident );
    DoPushConst( 1 );
    CurrentIdentScope.WriteMsg( 'add;' );
    DoPop( ident );
    DoPush( ident );
    end;
  CSTI_Decrement :
    begin
    CurrentIdentScope.WriteBeginStatusMsg( OutputLine, '-- ' );
    Scanner.Match( CSTI_Decrement );
    AssignmentHelper( OutputLine, Ident );
    DoPush( ident );
    DoPushConst( 1 );
    CurrentIdentScope.WriteMsg( 'sub;' );
    DoPop( ident );
    DoPush( ident );
    end;
  CSTI_OpenRound :
    begin
    CurrentIdentScope.WritePartialStatusMsg(OutputLine, '( ');
    Scanner.Match( CSTI_OpenRound );
    DoExpression( OutputLine );
    CurrentIdentScope.WritePartialStatusMsg(OutputLine, ') ');    
    Scanner.Match( CSTI_CloseRound );
    end;
  CSTI_Identifier :
    begin
    CurrentIdentScope.WritePartialStatusMsg(OutputLine, Scanner.Value + ' ');
    Ident := FindIdent( Scanner.Value, IdentNameIsScope );
    if Ident = nil then
       begin
       if not IdentNameIsScope then
         begin // referencing a non-existant variable
         WriteErrorMsg( NormalError, 'Undeclaired identifier, ' + Scanner.Value );
         Ident := NewIdent( Scanner.Value, UnknownVarType );
         end
       else  // inlined function call
         DoCallScript( true, OutputLine, true );
       end;
    if not IdentNameIsScope then
      begin
      DoPush( Ident );
      Scanner.GetToken;
      case Scanner.token of
        CSTI_Increment :
          begin
          CurrentIdentScope.WritePartialStatusMsg( OutputLine, '++ ' );
          Scanner.Match( CSTI_Increment );
          DoPush( Ident );
          DoPushConst( 1 );
          CurrentIdentScope.WriteMsg( 'add;' );
          DoPop( Ident );
          end;
        CSTI_Decrement :
          begin
          CurrentIdentScope.WriteBeginStatusMsg( OutputLine, '-- ' );
          Scanner.Match( CSTI_Decrement );
          DoPush( Ident );
          DoPushConst( 1 );
          CurrentIdentScope.WriteMsg( 'sub;' );
          DoPop( Ident );
          end;
      end;    
      end;
    end;
  CSTI_Integer, CSTI_HexInt :
    begin
    CurrentIdentScope.WritePartialStatusMsg(OutputLine, Scanner.Value + ' ');
    IntValue := StrToInt64( Scanner.Value );
    if IntValue > High(Longword) then
      IntValue := IntValue mod High(Longword);
    DoPushConst( IntValue );
    Scanner.GetToken;
    end;
  CSTI_Real :
    begin
    CurrentIdentScope.WritePartialStatusMsg(OutputLine, Scanner.Value + ' ');
    WriteErrorMsg( NormalError, 'Floatpoint values are not permitted in arbitrary locations ' + Scanner.Value );
    CurrentIdentScope.WriteMsg( 'pushc ' + Scanner.Value + ';' );
    Scanner.GetToken;
    end;
end;    
end; {DoFactor}

Procedure TBOSCompiler.DoUnaryFactor( OutputLine : integer );
var
  Token : TToken;
begin
while true do
  begin
  Token := Scanner.Token;
  if Token in [ CSTI_BitwiseNot, CSTI_LogicalNot, CSTI_Minus ] then
    begin
    CurrentIdentScope.WritePartialStatusMsg(OutputLine, TokenName( Scanner.Token )+ ' ');
    Scanner.GetToken;
    end;
  DoFactor( OutputLine );
  case Token of
    CSTI_BitwiseNot :
      begin
      if ScriptorCompatibility then
        CurrentIdentScope.WriteMsg( 'neg;' )
      else
        CurrentIdentScope.WriteMsg( 'not;' );
      end;
    CSTI_LogicalNot :
      begin
      CurrentIdentScope.WriteMsg( 'neg;' );
      end;
    CSTI_Minus :
      begin
      DoPushConst( 0 );
      CurrentIdentScope.WriteMsg( 'sub;' );
      end;
  else break;
  end;
  end;
end; {DoUnaryFactor}

Procedure TBOSCompiler.DoMulTerm( OutputLine : integer );
begin
DoUnaryFactor( OutputLine );
while true do
case Scanner.Token of
  CSTI_Multiply :
    begin
    CurrentIdentScope.WritePartialStatusMsg(OutputLine, '* ');
    Scanner.Match( CSTI_Multiply );
    DoUnaryFactor( OutputLine );
    CurrentIdentScope.WriteMsg( 'mul;' );
    end;
  CSTI_Divide :
    begin
    CurrentIdentScope.WritePartialStatusMsg(OutputLine, '\ ');
    Scanner.Match( CSTI_Divide );
    DoUnaryFactor( OutputLine );
    CurrentIdentScope.WriteMsg( 'div;' );
    end;
  else break;
end;
end; {DoMulTerm}

Procedure TBOSCompiler.DoAddTerm( OutputLine : integer );
begin
DoMulTerm( OutputLine );
while true do
case Scanner.Token of
  CSTI_Plus :
    begin
    CurrentIdentScope.WritePartialStatusMsg(OutputLine, '+ ');
    Scanner.Match( CSTI_Plus );
    DoMulTerm( OutputLine );
    CurrentIdentScope.WriteMsg( 'add;' );
    end;
  CSTI_Minus :
    begin
    CurrentIdentScope.WritePartialStatusMsg(OutputLine, '- ');
    Scanner.Match( CSTI_Minus );
    DoMulTerm( OutputLine );
    CurrentIdentScope.WriteMsg( 'sub;' );
    end;
  else break;
end;
end; {DoAddTerm}

Procedure TBOSCompiler.DoShiftTerm( OutputLine : integer );
begin
DoAddTerm( OutputLine );
if EnableBASMExtentions then
while true do
case Scanner.Token of
  CSTI_ShiftLeft :
    begin
    CurrentIdentScope.WritePartialStatusMsg(OutputLine, '<< ');
    Scanner.Match( CSTI_ShiftLeft );
    DoAddTerm( OutputLine );
    CurrentIdentScope.WriteMsg( 'shl;' );
    end;
  CSTI_ShiftRight :
    begin
    CurrentIdentScope.WritePartialStatusMsg(OutputLine, '>> ');
    Scanner.Match( CSTI_ShiftRight );
    DoAddTerm( OutputLine );
    CurrentIdentScope.WriteMsg( 'sar;' );
    end;
  else break;
end;
end; {DoShiftTerm}

Procedure TBOSCompiler.DoRelationTerm( OutputLine : integer );
begin
DoShiftTerm( OutputLine );
case Scanner.Token of
  CSTI_Greater :
    begin
    CurrentIdentScope.WritePartialStatusMsg(OutputLine, '> ');
    Scanner.Match( CSTI_Greater );
    DoShiftTerm( OutputLine );
    CurrentIdentScope.WriteMsg( 'setg;' );
    end;
  CSTI_GreaterEqual :
    begin
    CurrentIdentScope.WritePartialStatusMsg(OutputLine, '>= ');
    Scanner.Match( CSTI_GreaterEqual );
    DoShiftTerm( OutputLine );
    CurrentIdentScope.WriteMsg( 'setge;' );
    end;
  CSTI_Less :
    begin
    CurrentIdentScope.WritePartialStatusMsg(OutputLine, '< ');
    Scanner.Match( CSTI_Less );
    DoShiftTerm( OutputLine );
    CurrentIdentScope.WriteMsg( 'setl;' );
    end;
  CSTI_LessEqual :
    begin
    CurrentIdentScope.WritePartialStatusMsg(OutputLine, '<= ');
    Scanner.Match( CSTI_LessEqual );
    DoShiftTerm( OutputLine );
    CurrentIdentScope.WriteMsg( 'setle;' );
    end;
end;
end; {DoRelationTerm}

Procedure TBOSCompiler.DoEqualityTerm( OutputLine : integer );
begin
DoRelationTerm( OutputLine );
case Scanner.Token of
  CSTI_Equal :
    begin
    CurrentIdentScope.WritePartialStatusMsg(OutputLine, '== ');
    Scanner.Match( CSTI_Equal );
    DoRelationTerm( OutputLine );
    CurrentIdentScope.WriteMsg( 'cmp;' );
    end;
  CSTI_NotEqual :
    begin
    CurrentIdentScope.WritePartialStatusMsg(OutputLine, '!= ');
    Scanner.Match( CSTI_NotEqual );
    DoRelationTerm( OutputLine );
    CurrentIdentScope.WriteMsg( 'setne;' );
    end;
end;
end; {DoEqualityTerm}


//  if ScriptorCompatibility then


Procedure TBOSCompiler.DoBitwiseOr( OutputLine : integer );
begin
DoEqualityTerm( OutputLine );
while Scanner.IfMatch( CSTI_BitwiseOr ) do
   begin
   CurrentIdentScope.WritePartialStatusMsg( OutputLine, '| ' );
   DoEqualityTerm( OutputLine );
   CurrentIdentScope.WriteMsg( 'or;' );
   end;
end; {DoBitwiseOr}

Procedure TBOSCompiler.DoBitwiseXor( OutputLine : integer );
begin
DoBitwiseOr( OutputLine );
while Scanner.IfMatch( CSTI_BitwiseXor ) do
  begin
  CurrentIdentScope.WritePartialStatusMsg( OutputLine, '^ ' );
  DoBitwiseOr( OutputLine );
  CurrentIdentScope.WriteMsg( 'xor;' );
  end;
end; {DoBitwiseXor}

Procedure TBOSCompiler.DoBitwiseAnd( OutputLine : integer );
begin
DoBitwiseXor( OutputLine );
while Scanner.IfMatch( CSTI_BitwiseAnd ) do
  begin
  CurrentIdentScope.WritePartialStatusMsg( OutputLine, '& ' );
  DoBitwiseXor( OutputLine );
  CurrentIdentScope.WriteMsg( 'and;' );
  end;
end; {DoBitwiseAnd}

Procedure TBOSCompiler.DoLogicalOr( OutputLine : integer );
begin
if ScriptorCompatibility then
  begin
  DoBitwiseAnd( OutputLine );
  while Scanner.IfMatch( CSTI_LogicalOr ) do
    begin
    CurrentIdentScope.WritePartialStatusMsg( OutputLine, '|| ' );
    DoBitwiseAnd( OutputLine );
    CurrentIdentScope.WriteMsg( 'lor;' );
    end;
  end
else
  begin
  DoLogicalXor( OutputLine );
  While Scanner.IfMatch( CSTI_LogicalOr ) do
    begin
    CurrentIdentScope.WritePartialStatusMsg( OutputLine, '|| ' );
    DoLogicalXor( OutputLine );
    CurrentIdentScope.WriteMsg( 'lor;' );
    end;
  end;
end; {DoLogicalOr}

Procedure TBOSCompiler.DoLogicalXor( OutputLine : integer );
begin
if ScriptorCompatibility then
  begin
  DoLogicalOr( OutputLine );
  while Scanner.IfMatch( CSTI_LogicalXor ) do
    begin
    CurrentIdentScope.WritePartialStatusMsg( OutputLine, '^^ ' );
    DoLogicalOr( OutputLine );
    CurrentIdentScope.WriteMsg( 'lxor;' );
    end;
  end
else
  begin
  DoLogicalAnd( OutputLine );
  while Scanner.IfMatch( CSTI_LogicalXor ) do
    begin
    CurrentIdentScope.WritePartialStatusMsg( OutputLine, '^^ ' );
    DoLogicalAnd( OutputLine );
    CurrentIdentScope.WriteMsg( 'lxor;' );
    end;
  end;
end; {DoLogicalxor}

Procedure TBOSCompiler.DoLogicalAnd( OutputLine : integer );
begin
if ScriptorCompatibility then
  begin
  DoLogicalXor(OutputLine);
  while Scanner.IfMatch( CSTI_LogicalAnd ) do
    begin
    CurrentIdentScope.WritePartialStatusMsg( OutputLine, '&& ' );
    DoLogicalXor( OutputLine );
    CurrentIdentScope.WriteMsg( 'land;' );
    end;
  end
else
  begin
  DoBitwiseAnd( OutputLine );
  while Scanner.IfMatch( CSTI_LogicalAnd ) do
    begin
    CurrentIdentScope.WritePartialStatusMsg( OutputLine, '&& ' );
    DoBitwiseAnd( OutputLine );
    CurrentIdentScope.WriteMsg( 'land;' );
    end;
  end;
end; {DoLogicalAnd}

Procedure TBOSCompiler.DoExpression( OutputLine : integer );
begin
if ScriptorCompatibility then
  DoLogicalAnd( OutputLine )
else
  DoLogicalOr( OutputLine );
end; {DoExpression}

end.

