program BOS_to_BASM;
{$APPTYPE CONSOLE}
uses
  SysUtils,
  classes,
  inifiles,
  BOSCompilerU in 'BOSCompilerU.pas',
  PreProcessorU in 'PreProcessorU.pas',
  ScannerU in 'ScannerU.pas',
  Format_Cob in 'Format_Cob.pas',
  FileHandlingU in 'FileHandlingU.pas';


{$R *.res}
var
  EnableBASMExtentions : boolean;
  AssemblerExeName : string;
  AssemblerParams : string;
  AssemblerOutputFile  : string;
  CompileAndDoNotAssemble : boolean;
  PauseOnException : boolean;
  PauseAtEnd : boolean;
  AutoEmitLastRet : boolean;
  ScriptorCompatibility : Boolean;
Procedure ReadSettings;
var
  IniFile : TIniFile;
begin
IniFile := TIniFile.create(ChangeFileExt(ParamStr(0),'.ini'));
try
  IncludePaths := TStringlist.create;
  IncludePaths.Duplicates := DupIgnore;
  ReadSemiColonDelimitedList( IniFile.ReadString( 'Settings', 'IncludePaths', '' ), IncludePaths );
  SearchPaths := TStringlist.create;
  SearchPaths.Duplicates := DupIgnore;
  ReadSemiColonDelimitedList( IniFile.ReadString( 'Settings', 'SearchPaths', '' ), SearchPaths );
  SearchPaths.Add( ExtractFILePath( Paramstr( 0 ) ) );
  EnableBASMExtentions := IniFile.ReadBool( 'Settings', 'EnableBASMExtentions', true );
  AssemblerExeName := IniFile.ReadString( 'Settings', 'AssemblerExeName', 'BASM.exe' );
  AssemblerParams := IniFile.ReadString( 'Settings', 'AssemblerParams', '-l -o0 -t0' );
  AssemblerOutputFile := IniFile.ReadString( 'Settings', 'AssemblerOutputFile', '' );
  CompileAndDoNotAssemble := IniFile.ReadBool( 'Settings', 'CompileAndDoNotAssemble', false );
  PauseOnException := IniFile.ReadBool( 'Settings', 'PauseOnException', true );
  PauseAtEnd := IniFile.ReadBool( 'Settings', 'PauseAtEnd', false );
  AutoEmitLastRet := IniFile.ReadBool( 'Settings', 'AutoEmitLastRet', true );
  ScriptorCompatibility := IniFile.ReadBool( 'Settings', 'ScriptorCompatibility', true );
finally
  IniFile.free;
end;
end; {ReadSettings}

Procedure WriteSettings;
var
  IniFile : TIniFile;
begin
IniFile := TIniFile.create(ChangeFileExt(ParamStr(0),'.ini'));
try
//  ReadSemiColonDelimitedList( IniFile.ReadString( 'Settings', 'IncludePaths', '' ), IncludePaths );
//  ReadSemiColonDelimitedList( IniFile.ReadString( 'Settings', 'SearchPaths', '' ), SearchPaths );
  IniFile.WriteBool( 'Settings', 'EnableBASMExtentions', EnableBASMExtentions );
  IniFile.WriteString( 'Settings', 'AssemblerExeName', AssemblerExeName );
  IniFile.WriteString( 'Settings', 'AssemblerParams', AssemblerParams );
  IniFile.WriteString( 'Settings', 'AssemblerOutputFile', AssemblerOutputFile );
  IniFile.WriteBool( 'Settings', 'CompileAndDoNotAssemble', CompileAndDoNotAssemble );
  IniFile.WriteBool( 'Settings', 'PauseOnException', PauseOnException );
  IniFile.WriteBool( 'Settings', 'PauseAtEnd', PauseAtEnd );
  IniFile.WriteBool( 'Settings', 'AutoEmitLastRet', AutoEmitLastRet );
  IniFile.WriteBool( 'Settings', 'ScriptorCompatibility', ScriptorCompatibility );
finally
  IniFile.Updatefile;
  IniFile.free;
end;
end; {ReadSettings}

var
  Filename,FileNameRootPath : string;
  index : integer;
begin
  ThousandSeparator := ',';
  DecimalSeparator := '.';
try
  if ParamCount < 1 then
    Raise Exception.createfmt( 'Usage: %s <bos file>', [ ExtractFilename( ParamStr( 0 ) ) ] );
  SetCurrentDir( ExtractFilePath( ParamStr( 0 ) ) );
  ReadSettings;
  WriteSettings;
  for index := 1 to ParamCount do
    try try
      Filename := FindFile( ParamStr( index ) );
      if not FileExists(Filename) then
        Raise Exception.createfmt( 'File %s does not exist', [ Filename ] );
      Filename := ExtractRelativePath( IncludeTrailingBackslash( GetCurrentDir ), Filename);
      FileNameRootPath := ExtractFilePath( Filename );
      if FileNameRootPath <> '' then
        SearchPaths.Insert( 0, FileNameRootPath );
      CompileBOS( Filename, EnableBASMExtentions , CompileAndDoNotAssemble, AutoEmitLastRet,
                  ScriptorCompatibility,
                  AssemblerExeName, AssemblerParams, AssemblerOutputFile );
    finally
      Writeln;
      FreeAndNil( IncludePaths );
      FreeAndNil( SearchPaths );
    end;
    except
      On e : Exception do
        begin
        writeln( e.message );
        if PauseOnException and not PauseAtEnd then
          readln;
        end;
    end;
except
  On e : Exception do
    begin
    writeln( e.message );
    if PauseOnException and not PauseAtEnd then
      readln;
    end;
end;
if PauseAtEnd then
  readln;
end.