unit Format_Cob;

interface
uses
  classes,
  contnrs;
{$MINENUMSIZE 1}
type
// Every COB begins with this header
  PCobHeader = ^TCobHeader;
  TCobHeader = packed record
    // Specifies the COB version
    Version : longword;
    // The number of script Procs in this COB file
    NumberOfProcs : longword;
    // The number of pieces declared in this COB file
    NumberOfPieces : Longword;
    // The size in bytes of all the script Procs
    LengthOfAllProcs : longword;
    // The number of static variables declared in this COB
    NumberOfStaticVars : longword;
    // This is always 0
    Always_0 : longword;
    // Offset to an array of indecies. These indecies point to the script
    // Procs of the COB. The array consists of UINT32[ NumberOfProcs ]
    OffsetToProcPointerArray : longword;
    // Offset to an array of indecies. These indecies point to the names of
    // the script Procs in the name array. The array consists of UINT32[ NumberOfScripts ]
    OffsetToProcNameOffsetArray : longword;
    // Offset to an array of indecies. These indecies point to the names of
    // the pieces in the name array. The array consists of UINT32[ NumberOfPieces ]
    OffsetToPieceNameOffsetArray : longword;
    // Offset to the first Proc in the COB
    OffsetToFirstProc : LongInt;
    // Offset to the name array
    OffsetToNameArray : longword;
  end; {TTACobHeader}

  PTAKCobHeader = ^TTAKCobHeader;
  TTAKCobHeader = packed record
    TACobHeader : TCobHeader;
    // Offset to an array of indecies. These indecies point to the names of
    // the sounds in the name array. The array consists of UINT32[ NumberOfSounds ]
    OffsetToSoundNameArray : longword;
    // The number of sounds used in the script
    NumberOfSounds : longword;
  end; {TTAKCobHeader}

const
  // Indicates that the COB is from Total Annihilation
  CobVersion_TotalAnnihilation = $00000004;
  // Indicates that the COB is from TA:Kingdoms
  CobVersion_Kingdoms = $00000006;
  // Indicates that the COB is an enhanced TA cob
  CobVersion_TotalAnnihilation_Cobalt = $00000007;
// Known opcodes
Const
  Opcode_MovePieceWithSpeed           =  $10001000;
  Opcode_TurnPieceWithSpeed           =  $10002000;
  Opcode_StartSpin                    =  $10003000;
  Opcode_StopSpin                     =  $10004000;
  Opcode_ShowPiece                    =  $10005000;
  Opcode_HidePiece                    =  $10006000;
  Opcode_CachePiece                   =  $10007000;
  Opcode_DontCachePiece               =  $10008000;
  Opcode_DontShadow                   =  $1000A000;
  Opcode_MovePieceNow                 =  $1000B000;
  Opcode_TurnPieceNow                 =  $1000C000;
  Opcode_DontShade                    =  $1000E000;
  Opcode_EmitSfx                      =  $1000F000;
  Opcode_WaitForTurn                  =  $10011000;
  Opcode_WaitForMove                  =  $10012000;
  Opcode_Sleep                        =  $10013000;
  Opcode_PushConstantOntoStack        =  $10021001;
  Opcode_PushLocalVariableOntoStack   =  $10021002;
  Opcode_PushStaticVariableOntoStack  =  $10021004;
  Opcode_StackAllocate                =  $10022000;
  Opcode_SetLocalVariable             =  $10023002;
  Opcode_SetStaticVariable            =  $10023004;
  Opcode_PopStack                     =  $10024000;
  Opcode_Add                          =  $10031000;
  Opcode_Subtract                     =  $10032000;
  Opcode_Multiply                     =  $10033000;
  Opcode_Divide                       =  $10034000;
  Opcode_BitwiseAnd                   =  $10035000;
  Opcode_BitwiseOr                    =  $10036000;
  Opcode_Unknown1                     =  $10039000;
  Opcode_Unknown2                     =  $1003A000;
  Opcode_Unknown3                     =  $1003B000;
  Opcode_Random                       =  $10041000;
  Opcode_GetUnitValue                 =  $10042000;
  Opcode_GetFunctionResult            =  $10043000;
  Opcode_LessThan                     =  $10051000;
  Opcode_LessThanOrEqual              =  $10052000;
  Opcode_GreaterThan                  =  $10053000;
  Opcode_GreaterThanOrEqual           =  $10054000;
  Opcode_Equal                        =  $10055000;
  Opcode_NotEqual                     =  $10056000;
  Opcode_LogicalAnd                   =  $10057000;
  Opcode_LogicalOr                    =  $10058000;
  Opcode_LogicalNot                   =  $1005A000;
  Opcode_StartScript                  =  $10061000;
  Opcode_CallScript                   =  $10062000;
  Opcode_JumpToOffset                 =  $10064000;
  Opcode_Return                       =  $10065000;
  Opcode_JumpToOffsetIfFalse          =  $10066000;
  Opcode_Signal                       =  $10067000;
  Opcode_SetSignalMask                =  $10068000;
  Opcode_Explode                      =  $10071000;
  Opcode_PlaySound                    =  $10072000;
  Opcode_MapCommand                   =  $10073000;
  Opcode_SetUnitValue                 =  $10082000;
  Opcode_AttachUnit                   =  $10083000;
  Opcode_DropUnit                     =  $10084000;
  // Cobbler signature - First 4 bytes are 'Buil'; continues for  bytes
  Opcode_CobblerCrap                  =  $6c697542;
  // Fake opcodes used by Scriptor
  Opcode_ControlBlock                 =  $90000000;


type
  TUnitPropertyFlags = set of (SetAble, ArgIsPiece, ArgIsUnitID, IsBool);
  TUnitPropertyArgCount = 0..4;
  TUnitProperty = class
    fKey : Longint;
    fName : string;
    fFlags : TUnitPropertyFlags;
    fArgCount : TUnitPropertyArgCount;
  public
    constructor create( aKey : Longint;
                        const aName : string;
                        aFlags : TUnitPropertyFlags;
                        aArgCount : TUnitPropertyArgCount );

    Property Key : Longint read fKey;
    Property Name : string read fName;
    Property Flags : TUnitPropertyFlags read fFlags;
    // max of 5
    Property ArgCount : TUnitPropertyArgCount read fArgCount;
  end; {TUnitProperty}

  TUnitProperties = class
  protected
    flist : TObjectList;
    function GetUnitProperty( key : integer ) : TUnitProperty;
  public
    Constructor Create;
    Destructor Destroy; override;

    Procedure AddProperty( Key : Longint;
                           const Name : string;
                           Flags : TUnitPropertyFlags;
                           // max of 5
                           ArgCount : TUnitPropertyArgCount );
    Function ValidKey( key : integer) : boolean;
    Property UnitProperty[ key : integer] : TUnitProperty read GetUnitProperty; default;
  end; {TUnitProperties}

  TTAUnitProperties = class(TUnitProperties)
  public
    constructor create;
  end; {TTAUnitProperties}

  TTAKUnitProperties = class(TTAUnitProperties)
  public
    constructor create;
  end; {TTAKUnitProperties}

// Indices for set/get value
const
  // set Or get
  Activation         =  1;
  // set Or get
  StandingMoveOrders =  2;
  // set Or get
  StandingFireOrders =  3;
  // get (0-100%)
  Health             =  4;
  // set Or get
  InBuildStance      =  5;
  // set Or get (used by misc. special case missions like transport ships)
  Busy               =  6;
  // get
  Piece_XZ           =  7;
  // get
  Piece_Y            =  8;
  // get
  Unit_XZ            =  9;
  // get
  Unit_Y             = 10;
  // get
  Unit_Height        = 11;
  // get atan of packed x,z coords
  XZ_aTan            = 12;
  // get hypot of packed x,z coords
  XZ_Hypot           = 13;
  // get ordinary two-parameter atan
  aTan               = 14;
  // get ordinary two-parameter hypot
  Hypot              = 15;
  // get
  Ground_Height      = 16;
  // get 0 = unit is built and ready, 1-100 = How much is left to build
  Build_Percent_Left = 17;
  // set Or get (change which plots we occupy when building opens and closes)
  Yard_Open          = 18;
  // set Or get (ask other units to clear the area)
  Bugger_Off         = 19;
  // set Or get
  Armored            = 20;
// new in TA:K
  // setable
  Weapon_Aim_Aborted = 21;
  // setable
  Weapon_Ready       = 22;
  // setable
  Weapon_Launch_Now  = 23;
  // get & setable (bool)
  Finished_Dying     = 26;
  // get ?
  Orientation        = 27;
  // get (bool)
  In_Water           = 28;
  // get
  Current_Speed      = 29;
  // get (bool)
  Magic_Death        = 31;
  // get
  Veteran_Level      = 32;
  // get (bool)
  On_Road            = 34;

// TA:K maps  


// map script thingy
  UnKnownMapScriptFunctionCall = 30;

implementation
uses
  sysutils;
//------------------------------------------------------------------------------
// TUnitProperty
//------------------------------------------------------------------------------

constructor TUnitProperty.create( aKey : Longint;
                                  const aName : string;
                                  aFlags : TUnitPropertyFlags;
                                  aArgCount : TUnitPropertyArgCount );
begin
fKey := aKey;
fName := aName;
fFlags := aFlags;
fArgCount := aArgCount;
end; {create}

//------------------------------------------------------------------------------
// TUnitProperties
//------------------------------------------------------------------------------

Constructor TUnitProperties.Create;
begin
inherited create;
flist := TObjectList.create(true);
end; {Create}

Destructor TUnitProperties.Destroy;
begin
FreeAndNil( flist );
inherited;
end; {Destroy}

Procedure TUnitProperties.AddProperty( Key : Longint;
                                       const Name : string;
                                       Flags : TUnitPropertyFlags;
                                       ArgCount : TUnitPropertyArgCount );
var
  UnitProperty : TUnitProperty;
begin
if not ValidKey( Key ) then
  begin
  UnitProperty := TUnitProperty.create(Key, Name, Flags, ArgCount);
  try
    flist.add( UnitProperty );
  except
    UnitProperty.free;
    raise;
  end;
  end;
end; {AddProperty}

Function TUnitProperties.ValidKey( key : integer ) : boolean;
var
  index : integer;
begin
if flist <> nil then
for index := 0 to flist.count-1 do
  if TUnitProperty(flist[index]).Key = Key then
    begin
    result := true;
    exit;
    end;
result := false;
end; {ValidIndex}

function TUnitProperties.GetUnitProperty( key : integer ) : TUnitProperty;
var
  index : integer;
begin
if flist <> nil then
for index := 0 to flist.count-1 do
  begin
  result := TUnitProperty(flist[index]);
  if result.Key = Key then
    exit;
  end;
result := nil;
end; {GetUnitProperty}


//------------------------------------------------------------------------------
// TTAUnitProperties
//------------------------------------------------------------------------------

Constructor TTAUnitProperties.Create;
begin
inherited create;
AddProperty( Activation, 'Activation', [ SetAble,IsBool ], 0);
AddProperty( StandingMoveOrders, 'StandingMoveOrders', [ SetAble ], 0);
AddProperty( StandingFireOrders, 'StandingFireOrders', [ SetAble ], 0);
AddProperty( Health, 'Health', [ ], 0 );
AddProperty( InBuildStance, 'InBuildStance', [ SetAble ], 0);
AddProperty( Busy, 'Busy', [ SetAble, IsBool ], 0);
AddProperty( Piece_XZ, 'Piece_XZ', [ ArgIsPiece ], 1 );
AddProperty( Piece_Y, 'Piece_Y', [ ArgIsPiece ], 1 );
AddProperty( Unit_XZ, 'Unit_XZ', [ ArgIsUnitID ], 1 );
AddProperty( Unit_Y, 'Unit_Y', [ ArgIsUnitID ], 1 );
AddProperty( Unit_Height, 'Unit_Height', [ ArgIsUnitID ], 1 );
AddProperty( XZ_aTan, 'XZ_aTan', [ ], 1 );
AddProperty( XZ_Hypot, 'XZ_Hypot', [ ], 1 );
AddProperty( aTan, 'aTan', [ ], 2 );
AddProperty( Hypot, 'Hypot', [ ], 2 );
AddProperty( Ground_Height, 'Ground_Height', [ ], 1 );
AddProperty( Build_Percent_Left, 'Build_Percent_Left', [ ], 0 );
AddProperty( Yard_Open, 'Yard_Open', [ SetAble, IsBool ], 0 );
AddProperty( Bugger_Off, 'Bugger_Off', [ SetAble, IsBool ], 0 );
AddProperty( Armored, 'Armored', [ SetAble, IsBool ], 0 );
end; {Create}


//------------------------------------------------------------------------------
// TTAKUnitProperties
//------------------------------------------------------------------------------

Constructor TTAKUnitProperties.Create;
begin
inherited create;
AddProperty( Weapon_Aim_Aborted, 'Weapon_Aim_Aborted', [ Setable ], 0 );
AddProperty( Weapon_Ready, 'Weapon_Ready', [ Setable ], 0 );
AddProperty( Weapon_Launch_Now, 'Weapon_Launch_Now', [ Setable ], 0 );
AddProperty( Finished_Dying, 'Finished_Dying', [ Setable, IsBool ], 0 );
//AddProperty( Orientation, 'Orientation', [] , 0 );
AddProperty( In_Water, 'In_Water', [ IsBool ], 0 );
AddProperty( Current_Speed, 'Current_Speed', [] , 0 );  // ??
AddProperty( Magic_Death, 'Magic_Death', [ IsBool ], 0 );
AddProperty( Veteran_Level, 'Veteran_Level', [] , 0 );
AddProperty( On_Road, 'On_Road', [ IsBool ], 0 );
end; {Create}

end.
